# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2021
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
Implementation of the adversarial texture attack on object trackers in PyTorch.

| Paper link: https://arxiv.org/abs/1904.11042
"""
import logging
from typing import Dict, List, Optional, Union, TYPE_CHECKING

import numpy as np
from tqdm.auto import trange

from art.attacks.attack import EvasionAttack
from art.estimators.estimator import BaseEstimator, LossGradientsMixin
from art.estimators.object_tracking.object_tracker import ObjectTrackerMixin
from art.summary_writer import SummaryWriter

if TYPE_CHECKING:
    # pylint: disable=C0412
    import torch

logger = logging.getLogger(__name__)


class AdversarialTexturePyTorch(EvasionAttack):
    """
    Implementation of the adversarial texture attack on object trackers in PyTorch.

    | Paper link: https://arxiv.org/abs/1904.11042
    """

    attack_params = EvasionAttack.attack_params + [
        "patch_height",
        "patch_width",
        "xmin",
        "ymin",
        "step_size",
        "max_iter",
        "batch_size",
        "verbose",
    ]

    _estimator_requirements = (ObjectTrackerMixin, LossGradientsMixin, BaseEstimator)

    def __init__(
        self,
        estimator,
        patch_height: int,
        patch_width: int,
        x_min: int = 0,
        y_min: int = 0,
        step_size: float = 1.0 / 255.0,
        max_iter: int = 500,
        batch_size: int = 16,
        summary_writer: Union[str, bool, SummaryWriter] = False,
        verbose: bool = True,
    ):
        """
        Create an instance of the :class:`.AdversarialTexturePyTorch`.

        :param estimator: A trained estimator.
        :param patch_height: Height of patch.
        :param patch_width: Width of patch.
        :param x_min: Vertical position of patch, top-left corner.
        :param y_min: Horizontal position of patch, top-left corner.
        :param step_size: The step size.
        :param max_iter: The number of optimization steps.
        :param batch_size: The size of the training batch.
        :param summary_writer: Activate summary writer for TensorBoard.
                               Default is `False` and deactivated summary writer.
                               If `True` save runs/CURRENT_DATETIME_HOSTNAME in current directory.
                               If of type `str` save in path.
                               If of type `SummaryWriter` apply provided custom summary writer.
                               Use hierarchical folder structure to compare between runs easily. e.g. pass in
                               ‘runs/exp1’, ‘runs/exp2’, etc. for each new experiment to compare across them.
        :param verbose: Show progress bars.
        """
        import torch

        super().__init__(estimator=estimator, summary_writer=summary_writer)
        self.patch_height = patch_height
        self.patch_width = patch_width
        self.x_min = x_min
        self.y_min = y_min
        self.step_size = step_size
        self.max_iter = max_iter
        self.batch_size = batch_size
        self.verbose = verbose
        self._check_params()

        self._batch_id = 0
        self._i_max_iter = 0

        self.patch_shape = (self.patch_height, self.patch_width, 3)

        if self.estimator.channels_first:
            raise ValueError("Input shape has to be either NHWC or NFHWC.")

        self.i_h_patch = 0
        self.i_w_patch = 1

        self.i_h = 1
        self.i_w = 2

        if not (self.estimator.postprocessing_defences is None or self.estimator.postprocessing_defences == []):
            raise ValueError(
                "Framework-specific implementation of Adversarial Patch attack does not yet support "
                + "postprocessing defences."
            )

        mean_value = (self.estimator.clip_values[1] - self.estimator.clip_values[0]) / 2.0 + self.estimator.clip_values[
            0
        ]
        self._initial_value = np.ones(self.patch_shape) * mean_value
        self._patch = torch.tensor(self._initial_value, requires_grad=True, device=self.estimator.device)

    def _train_step(
        self,
        videos: "torch.Tensor",
        target: List[Dict[str, "torch.Tensor"]],
        y_init: "torch.Tensor",
        foreground: Optional["torch.Tensor"],
        patch_points: Optional[np.ndarray],
    ) -> "torch.Tensor":
        """
        Apply a training step to the batch based on a mini-batch.

        :param videos: Video samples.
        :param target: Target labels/boxes.
        :param y_init: Initial labels/boxes.
        :param foreground: Foreground mask.
        :param patch_points: Array of shape (nb_frames, 4, 2) containing four pairs of integers (height, width)
                             corresponding to the coordinates of the four corners top-left, top-right, bottom-right,
                             bottom-left of the transformed image in the coordinate-system of the original image.
        :return: Loss.
        """
        import torch

        self.estimator.model.zero_grad()
        loss = self._loss(videos, target, y_init, foreground, patch_points)
        loss.backward(retain_graph=True)

        if self._patch.grad is not None:
            gradients = self._patch.grad.sign() * self.step_size
        else:
            raise ValueError("Gradient term in PyTorch model is `None`.")

        # Write summary
        if self.summary_writer is not None:  # pragma: no cover
            self.summary_writer.update(
                batch_id=self._batch_id,
                global_step=self._i_max_iter,
                grad=np.expand_dims(self._patch.grad.detach().cpu().numpy(), axis=0),
                patch=None,
                estimator=None,
                x=None,
                y=None,
            )

        self._patch.grad = torch.zeros(
            self._patch.grad.shape, device=self._patch.grad.device, dtype=self._patch.grad.dtype
        )

        with torch.no_grad():
            self._patch[:] = torch.clamp(
                self._patch + gradients, min=self.estimator.clip_values[0], max=self.estimator.clip_values[1]
            )

        return loss

    def _predictions(
        self,
        videos: "torch.Tensor",
        y_init: "torch.Tensor",
        foreground: Optional["torch.Tensor"],
        patch_points: Optional[np.ndarray],
    ) -> List[Dict[str, "torch.Tensor"]]:
        """
        Predict object tracking estimator on patched videos.

        :param videos: Video samples.
        :param y_init: Initial labels/boxes.
        :param foreground: Foreground mask.
        :param patch_points: Array of shape (nb_frames, 4, 2) containing four pairs of integers (height, width)
                             corresponding to the coordinates of the four corners top-left, top-right, bottom-right,
                             bottom-left of the transformed image in the coordinate-system of the original image.
        :return: Predicted labels/boxes.
        """
        import torch

        patched_input = self._apply_texture(videos, self._patch, foreground=foreground, patch_points=patch_points)
        patched_input = torch.clamp(
            patched_input,
            min=self.estimator.clip_values[0],
            max=self.estimator.clip_values[1],
        )

        predictions = self.estimator.predict(patched_input, y_init=y_init)  # pylint: disable=W0212

        return predictions

    def _loss(
        self,
        videos: "torch.Tensor",
        target: List[Dict[str, "torch.Tensor"]],
        y_init: "torch.Tensor",
        foreground: Optional["torch.Tensor"],
        patch_points: Optional[np.ndarray],
    ) -> "torch.Tensor":
        """
        Calculate L1-loss.

        :param videos: Video samples.
        :param target: Target labels/boxes.
        :param y_init: Initial labels/boxes.
        :param foreground: Foreground mask.
        :param patch_points: Array of shape (nb_frames, 4, 2) containing four pairs of integers (height, width)
                             corresponding to the coordinates of the four corners top-left, top-right, bottom-right,
                             bottom-left of the transformed image in the coordinate-system of the original image.
        :return: Loss.
        """
        import torch

        y_pred = self._predictions(videos, y_init, foreground, patch_points)
        loss = torch.nn.L1Loss(reduction="sum")(y_pred[0]["boxes"].float(), target[0]["boxes"].float())
        for i in range(1, len(y_pred)):
            loss = loss + torch.nn.L1Loss(reduction="sum")(y_pred[i]["boxes"].float(), target[i]["boxes"].float())

        return loss

    def _get_patch_mask(self, nb_samples: int) -> "torch.Tensor":
        """
        Create patch mask.

        :param nb_samples: Number of samples.
        :return: Patch mask.
        """
        import torch

        image_mask_np = np.ones((self.patch_height, self.patch_width))

        image_mask_np = np.expand_dims(image_mask_np, axis=2)
        image_mask_np = np.broadcast_to(image_mask_np, self.patch_shape)
        image_mask = torch.Tensor(np.array(image_mask_np)).to(self.estimator.device)
        image_mask = torch.stack([image_mask] * nb_samples, dim=0)
        return image_mask

    def _apply_texture(
        self,
        videos: "torch.Tensor",
        patch: "torch.Tensor",
        foreground: Optional["torch.Tensor"],
        patch_points: Optional[np.ndarray],
    ) -> "torch.Tensor":
        """
        Apply texture over background and overlay foreground.

        :param videos: Video samples.
        :param patch: Patch to apply.
        :param foreground: Foreground mask.
        :param patch_points: Array of shape (nb_frames, 4, 2) containing four pairs of integers (height, width)
                             corresponding to the coordinates of the four corners top-left, top-right, bottom-right,
                             bottom-left of the transformed image in the coordinate-system of the original image.
        :return: Patched videos.
        """
        import torch
        import torchvision

        nb_samples = videos.shape[0]
        nb_frames = videos.shape[1]
        frame_height = videos.shape[2]
        frame_width = videos.shape[3]

        image_mask = self._get_patch_mask(nb_samples=nb_samples)
        image_mask = image_mask.float()

        patch = patch.float()
        padded_patch = torch.stack([patch] * nb_samples)

        if patch_points is None:
            pad_h_before = self.x_min
            pad_h_after = int(videos.shape[self.i_h + 1] - pad_h_before - image_mask.shape[self.i_h_patch + 1])

            pad_w_before = self.y_min
            pad_w_after = int(videos.shape[self.i_w + 1] - pad_w_before - image_mask.shape[self.i_w_patch + 1])

            image_mask = image_mask.permute(0, 3, 1, 2)

            image_mask = torchvision.transforms.functional.pad(
                img=image_mask,
                padding=[pad_w_before, pad_h_before, pad_w_after, pad_h_after],
                fill=0,
                padding_mode="constant",
            )

            image_mask = image_mask.permute(0, 2, 3, 1)

            image_mask = torch.unsqueeze(image_mask, dim=1)
            image_mask = torch.repeat_interleave(image_mask, dim=1, repeats=nb_frames)
            image_mask = image_mask.float()

            padded_patch = padded_patch.permute(0, 3, 1, 2)

            padded_patch = torchvision.transforms.functional.pad(
                img=padded_patch,
                padding=[pad_w_before, pad_h_before, pad_w_after, pad_h_after],
                fill=0,
                padding_mode="constant",
            )

            padded_patch = padded_patch.permute(0, 2, 3, 1)

            padded_patch = torch.unsqueeze(padded_patch, dim=1)
            padded_patch = torch.repeat_interleave(padded_patch, dim=1, repeats=nb_frames)

            padded_patch = padded_patch.float()

        else:

            startpoints = [[0, 0], [frame_width, 0], [frame_width, frame_height], [0, frame_height]]
            endpoints = np.zeros_like(patch_points)
            endpoints[:, :, 0] = patch_points[:, :, 1]
            endpoints[:, :, 1] = patch_points[:, :, 0]

            image_mask = image_mask.permute(0, 3, 1, 2)

            image_mask = torchvision.transforms.functional.resize(
                img=image_mask,
                size=[int(videos.shape[2]), int(videos.shape[3])],
                interpolation=torchvision.transforms.InterpolationMode.BILINEAR,
            )

            image_mask_list = []

            for i_frame in range(nb_frames):

                image_mask_i = torchvision.transforms.functional.perspective(
                    img=image_mask,
                    startpoints=startpoints,
                    endpoints=endpoints[i_frame],
                    interpolation=torchvision.transforms.InterpolationMode.BILINEAR,
                    fill=0,
                )

                image_mask_i = image_mask_i.permute(0, 2, 3, 1)

                image_mask_list.append(image_mask_i)

            image_mask = torch.stack(image_mask_list, dim=1)
            image_mask = image_mask.float()

            padded_patch = padded_patch.permute(0, 3, 1, 2)

            padded_patch = torchvision.transforms.functional.resize(
                img=padded_patch,
                size=[int(videos.shape[2]), int(videos.shape[3])],
                interpolation=torchvision.transforms.InterpolationMode.BILINEAR,
            )

            padded_patch_list = []

            for i_frame in range(nb_frames):
                padded_patch_i = torchvision.transforms.functional.perspective(
                    img=padded_patch,
                    startpoints=startpoints,
                    endpoints=endpoints[i_frame],
                    interpolation=torchvision.transforms.InterpolationMode.BILINEAR,
                    fill=0,
                )

                padded_patch_i = padded_patch_i.permute(0, 2, 3, 1)

                padded_patch_list.append(padded_patch_i)

            padded_patch = torch.stack(padded_patch_list, dim=1)
            padded_patch = padded_patch.float()

        inverted_mask = (
            torch.from_numpy(np.ones(shape=image_mask.shape, dtype=np.float32)).to(self.estimator.device) - image_mask
        )

        if foreground is not None:
            combined = (
                videos * inverted_mask
                + padded_patch * image_mask
                - padded_patch * ~foreground.bool()
                + videos * ~foreground.bool() * image_mask
            )
        else:
            combined = videos * inverted_mask + padded_patch * image_mask

        return combined

    def generate(  # type: ignore  # pylint: disable=W0222
        self, x: np.ndarray, y: List[Dict[str, np.ndarray]], **kwargs
    ) -> np.ndarray:
        """
        Generate an adversarial patch and return the patch and its mask in arrays.

        :param x: Input videos of shape NFHWC.
        :param y: True labels of format `List[Dict[str, np.ndarray]]`, one dictionary for each input image. The keys of
                  the dictionary are:

                  - boxes [N_FRAMES, 4]: the boxes in [x1, y1, x2, y2] format, with 0 <= x1 < x2 <= W and
                                         0 <= y1 < y2 <= H.

        :Keyword Arguments:
            * *shuffle* (``np.ndarray``) --
              Shuffle order of samples, labels, initial boxes, and foregrounds for texture generation.
            * *y_init* (``np.ndarray``) --
              Initial boxes around object to be tracked of shape (nb_samples, 4) with second dimension representing
              [x1, y1, x2, y2] with 0 <= x1 < x2 <= W and 0 <= y1 < y2 <= H.
            * *foreground* (``np.ndarray``) --
              Foreground masks of shape NFHWC of boolean values with False/0.0 representing foreground, preventing
              updates to the texture, and True/1.0 for background, allowing updates to the texture.
            * *patch_points* (``np.ndarray``) --
              Array of shape (nb_frames, 4, 2) containing four pairs of integers (height, width) corresponding to the
              four corners top-left, top-right, bottom-right, bottom-left of the transformed image in the coordinates
              of the original image.

        :return: An array with images patched with adversarial texture.
        """
        import torch

        shuffle = kwargs.get("shuffle", True)
        y_init = kwargs.get("y_init")
        foreground = kwargs.get("foreground")
        if foreground is None:
            foreground = np.ones_like(x)
        patch_points = kwargs.get("patch_points")

        class TrackingDataset(torch.utils.data.Dataset):
            """
            Object tracking dataset in PyTorch.
            """

            def __init__(self, x, y, y_init, foreground):
                self.x = x
                self.y = y
                self.y_init = y_init
                self.foreground = foreground

            def __len__(self):
                return self.x.shape[0]

            def __getitem__(self, idx):
                img = self.x[idx]

                target = {}
                target["boxes"] = torch.from_numpy(y[idx]["boxes"])

                y_init_i = self.y_init[idx]
                foreground_i = self.foreground[idx]

                return img, target, y_init_i, foreground_i

        dataset = TrackingDataset(x, y, y_init, foreground)
        data_loader = torch.utils.data.DataLoader(
            dataset=dataset,
            batch_size=self.batch_size,
            shuffle=shuffle,
            drop_last=False,
        )

        for i_max_iter in trange(self.max_iter, desc="Adversarial Texture PyTorch", disable=not self.verbose):

            self._i_max_iter = i_max_iter
            self._batch_id = 0

            for videos_i, target_i, y_init_i, foreground_i in data_loader:
                self._batch_id += 1
                videos_i = videos_i.to(self.estimator.device)
                y_init_i = y_init_i.to(self.estimator.device)
                foreground_i = foreground_i.to(self.estimator.device)
                target_i_list = []
                for i_t in range(videos_i.shape[0]):
                    target_i_list.append({"boxes": target_i["boxes"][i_t].to(self.estimator.device)})

                _ = self._train_step(
                    videos=videos_i,
                    target=target_i_list,
                    y_init=y_init_i,
                    foreground=foreground_i,
                    patch_points=patch_points,
                )

                # Write summary
                if self.summary_writer is not None:  # pragma: no cover
                    self.summary_writer.update(
                        batch_id=self._batch_id,
                        global_step=self._i_max_iter,
                        grad=None,
                        patch=self._patch.detach().cpu().numpy(),
                        estimator=self.estimator,
                        x=videos_i.detach().cpu().numpy(),
                        y=target_i_list,
                    )

        if self.summary_writer is not None:
            self.summary_writer.reset()

        return self.apply_patch(x=x, foreground=foreground, patch_points=patch_points)

    def apply_patch(
        self,
        x: np.ndarray,
        patch_external: Optional[np.ndarray] = None,
        foreground: Optional[np.ndarray] = None,
        patch_points: Optional[np.ndarray] = None,
    ) -> np.ndarray:
        """
        A function to apply the learned adversarial texture to videos.

        :param x: Videos of shape NFHWC to apply adversarial texture.
        :param patch_external: External patch to apply to videos `x`.
        :param foreground: Foreground masks of shape NFHWC of boolean values with False/0.0 representing foreground,
                           preventing updates to the texture, and True/1.0 for background, allowing updates to the
                           texture.
        :param patch_points: Array of shape (nb_frames, 4, 2) containing four pairs of integers (height, width)
                             corresponding to the coordinates of the four corners top-left, top-right, bottom-right,
                             bottom-left of the transformed image in the coordinate-system of the original image.
        :return: The videos with adversarial textures.
        """
        import torch

        patch_tensor = (
            torch.Tensor(patch_external).to(self.estimator.device) if patch_external is not None else self._patch
        )
        x_tensor = torch.Tensor(x).to(self.estimator.device)
        if foreground is None:
            foreground_tensor = None
        else:
            foreground_tensor = torch.Tensor(foreground).to(self.estimator.device)

        return (
            self._apply_texture(
                videos=x_tensor, patch=patch_tensor, foreground=foreground_tensor, patch_points=patch_points
            )
            .detach()
            .cpu()
            .numpy()
        )

    def reset_patch(self, initial_patch_value: Optional[Union[float, np.ndarray]] = None) -> None:
        """
        Reset the adversarial texture.

        :param initial_patch_value: Patch value to use for resetting the patch.
        """
        import torch

        if initial_patch_value is None:
            self._patch.data = torch.Tensor(self._initial_value).double()
        elif isinstance(initial_patch_value, float):
            initial_value = np.ones(self.patch_shape) * initial_patch_value
            self._patch.data = torch.Tensor(initial_value).double()
        elif self._patch.shape == initial_patch_value.shape:
            self._patch.data = torch.Tensor(initial_patch_value).double()
        else:
            raise ValueError("Unexpected value for initial_patch_value.")

    def _check_params(self) -> None:

        if not isinstance(self.patch_height, int) or self.patch_height <= 0:
            raise ValueError("The patch height `patch_height` has to be of type int and larger than zero.")

        if not isinstance(self.patch_width, int) or self.patch_width <= 0:
            raise ValueError("The patch width `patch_width` has to be of type int and larger than zero.")

        if not isinstance(self.x_min, int) or self.x_min < 0:
            raise ValueError("The vertical position `x_min` has to be of type int and larger than zero.")

        if not isinstance(self.y_min, int) or self.y_min < 0:
            raise ValueError("The horizontal position `y_min` has to be of type int and larger than zero.")

        if not isinstance(self.step_size, float) or self.step_size <= 0:
            raise ValueError("The step size `step_size` has to be of type float and larger than zero.")

        if not isinstance(self.max_iter, int) or self.max_iter <= 0:
            raise ValueError("The number of iterations `max_iter` has to be of type int and larger than zero.")

        if not isinstance(self.batch_size, int) or self.batch_size <= 0:
            raise ValueError("The batch size `batch_size` has to be of type int and larger than zero.")

        if not isinstance(self.verbose, bool):
            raise ValueError("The argument `verbose` has to be of type bool.")



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2023
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# MIT License
#
# Copyright (c) 2022 Chong Xiang
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This module implements the ObjectSeeker certifiably robust defense.

| Paper link: https://arxiv.org/abs/2202.01811
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import abc
import logging
from typing import Dict, List, Tuple

import numpy as np
from sklearn.cluster import DBSCAN
from tqdm.auto import tqdm

from art.utils import intersection_over_area

logger = logging.getLogger(__name__)


class ObjectSeekerMixin(abc.ABC):
    """
    Implementation of the ObjectSeeker certifiable robust defense applied to object detection models.
    The original implementation is https://github.com/inspire-group/ObjectSeeker

    | Paper link: https://arxiv.org/abs/2202.01811
    """

    def __init__(
        self,
        *args,
        num_lines: int = 3,
        confidence_threshold: float = 0.3,
        iou_threshold: float = 0.5,
        prune_threshold: float = 0.5,
        epsilon: float = 0.1,
        verbose: bool = False,
        **kwargs,
    ) -> None:
        """
        Create an ObjectSeeker wrapper.

        :param num_lines: The number of divisions both vertically and horizontally to make masked predictions.
        :param confidence_threshold: The confidence threshold to discard bounding boxes.
        :param iou_threshold: The IoU threshold to discard overlapping bounding boxes.
        :param prune_threshold: The IoA threshold for pruning and duplicated bounding boxes.
        :param epsilon: The maximum distance between bounding boxes when merging using DBSCAN.
        :param verbose: Show progress bars.
        """
        super().__init__(*args, **kwargs)  # type: ignore
        self.num_lines = num_lines
        self.confidence_threshold = confidence_threshold
        self.iou_threshold = iou_threshold
        self.prune_threshold = prune_threshold
        self.epsilon = epsilon
        self.verbose = verbose

    @abc.abstractmethod
    def _image_dimensions(self) -> Tuple[int, int]:
        """
        Get the height and width of a sample input image.

        :return: Tuple containing the height and width of a sample input image.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def _masked_predictions(
        self, x_i: np.ndarray, batch_size: int = 128, **kwargs
    ) -> Tuple[Dict[str, np.ndarray], Dict[str, np.ndarray]]:
        """
        Create masked copies of the image for each of lines following the ObjectSeeker algorithm. Then creates
        predictions on the base unmasked image and each of the masked image.

        :param x_i: A single image of shape CHW or HWC.
        :batch_size: Batch size.
        :return: Predictions for the base unmasked image and merged predictions for the masked image.
        """
        raise NotImplementedError

    def _prune_boxes(
        self, masked_preds: Dict[str, np.ndarray], base_preds: Dict[str, np.ndarray]
    ) -> Dict[str, np.ndarray]:
        """
        Remove bounding boxes from the masked predictions of a single image based on the IoA score with the boxes
        on the base unmasked predictions.

        :param masked_preds: The merged masked predictions of a single image.
        :param base_preds: The base unmasked predictions of a single image.
        :return: The filtered masked predictions with extraneous boxes removed.
        """
        masked_boxes = masked_preds["boxes"]
        masked_labels = masked_preds["labels"]
        masked_scores = masked_preds["scores"]

        base_boxes = base_preds["boxes"]
        base_labels = base_preds["labels"]

        keep_indices = []
        for idx, (masked_box, masked_label) in enumerate(zip(masked_boxes, masked_labels)):
            keep = True
            for (base_box, base_label) in zip(base_boxes, base_labels):
                if masked_label == base_label:
                    ioa = intersection_over_area(masked_box, base_box)
                    if ioa >= self.prune_threshold:
                        keep = False
                        break

            if keep:
                keep_indices.append(idx)

        pruned_preds = {
            "boxes": masked_boxes[keep_indices],
            "labels": masked_labels[keep_indices],
            "scores": masked_scores[keep_indices],
        }
        return pruned_preds

    def _unionize_clusters(self, masked_preds: Dict[str, np.ndarray]) -> Dict[str, np.ndarray]:
        """
        Cluster the bounding boxes for the pruned masked predictions.

        :param masked_preds: The merged masked predictions of a single image already pruned.
        :return: The clustered masked predictions with overlapping boxes merged.
        """
        boxes = masked_preds["boxes"]
        labels = masked_preds["labels"]
        scores = masked_preds["scores"]

        # Skip clustering if not enough bounding boxes
        if len(boxes) <= 1:
            return masked_preds

        unionized_boxes = []
        unionized_labels = []
        unionized_scores = []

        unique_labels = np.unique(labels)
        for label in unique_labels:
            # Select bounding boxes of same class
            mask = labels == label
            selected_boxes = boxes[mask]
            selected_scores = scores[mask]

            # Calculate pairwise distances between bounding boxes
            areas = (selected_boxes[:, 2] - selected_boxes[:, 0]) * (selected_boxes[:, 3] - selected_boxes[:, 1])
            top_left = np.maximum(selected_boxes[:, None, :2], selected_boxes[:, :2])
            bottom_right = np.minimum(selected_boxes[:, None, 2:], selected_boxes[:, 2:])
            pairwise_intersection = np.prod(np.clip(bottom_right - top_left, 0, None), axis=2)
            pairwise_ioa = pairwise_intersection / (areas[:, None])
            distances = 1 - np.maximum(pairwise_ioa, pairwise_ioa.T)

            # Cluster bounding boxes
            dbscan = DBSCAN(eps=self.epsilon, min_samples=1, metric="precomputed")
            clusters = dbscan.fit_predict(distances)
            num_clusters = np.max(clusters) + 1

            for cluster in range(num_clusters):
                # Merge all boxes in the cluster
                clustered_boxes = selected_boxes[clusters == cluster]

                clustered_box = [
                    np.min(clustered_boxes[:, 0]),
                    np.min(clustered_boxes[:, 1]),
                    np.max(clustered_boxes[:, 2]),
                    np.max(clustered_boxes[:, 3]),
                ]
                clustered_score = np.max(selected_scores)

                unionized_boxes.append(clustered_box)
                unionized_labels.append(label)
                unionized_scores.append(clustered_score)

        unionized_predictions = {
            "boxes": np.asarray(unionized_boxes),
            "labels": np.asarray(unionized_labels),
            "scores": np.asarray(unionized_scores),
        }
        return unionized_predictions

    def predict(self, x: np.ndarray, batch_size: int = 128, **kwargs) -> List[Dict[str, np.ndarray]]:
        """
        Perform prediction for a batch of inputs.

        :param x: Samples of shape NCHW or NHWC.
        :param batch_size: Batch size.
        :return: Predictions of format `List[Dict[str, np.ndarray]]`, one for each input image. The fields of the Dict
                 are as follows:

                 - boxes [N, 4]: the boxes in [x1, y1, x2, y2] format, with 0 <= x1 < x2 <= W and 0 <= y1 < y2 <= H.
                 - labels [N]: the labels for each image
                 - scores [N]: the scores or each prediction.
        """
        predictions = []

        for x_i in tqdm(x, desc="ObjectSeeker", disable=not self.verbose):
            base_preds, masked_preds = self._masked_predictions(x_i, batch_size=batch_size, **kwargs)
            pruned_preds = self._prune_boxes(masked_preds, base_preds)
            unionized_preds = self._unionize_clusters(pruned_preds)

            preds = {
                "boxes": np.concatenate([base_preds["boxes"], unionized_preds["boxes"]]),
                "labels": np.concatenate([base_preds["labels"], unionized_preds["labels"]]),
                "scores": np.concatenate([base_preds["scores"], unionized_preds["scores"]]),
            }

            predictions.append(preds)

        return predictions

    def certify(
        self,
        x: np.ndarray,
        patch_size: float = 0.01,
        offset: float = 0.1,
        batch_size: int = 128,
    ) -> List[np.ndarray]:
        """
        Checks if there is certifiable IoA robustness for each predicted bounding box.

        :param x: Sample input with shape as expected by the model.
        :param patch_size: The size of the patch to check against.
        :param offset: The offset to distinguish between the far and near patches.
        :return: A list containing an array of bools for each bounding box per image indicating if the bounding
                 box is certified against the given patch.
        """
        height, width = self._image_dimensions()

        patch_size = np.sqrt(height * width * patch_size)
        height_offset = offset * height
        width_offset = offset * width

        # Get predictions
        predictions = self.predict(x, batch_size=batch_size)

        certifications: List[np.ndarray] = []

        for pred in tqdm(predictions, desc="ObjectSeeker", disable=not self.verbose):
            boxes = pred["boxes"]

            # Far patch
            far_patch_map = np.ones((len(boxes), height, width), dtype=bool)
            for i, box in enumerate(boxes):
                x_1 = int(max(0, box[1] - patch_size - height_offset))
                x_2 = int(min(box[3] + height_offset + 1, height))
                y_1 = int(max(0, box[0] - patch_size - width_offset))
                y_2 = int(min(box[2] + width_offset + 1, width))
                far_patch_map[i, x_1:x_2, y_1:y_2] = False
            far_vulnerable = np.any(far_patch_map, axis=(-2, -1))

            # Close patch
            close_patch_map = np.ones((len(boxes), height, width), dtype=bool)
            for i, box in enumerate(boxes):
                x_1 = int(max(0, box[1] - patch_size))
                x_2 = int(min(box[3] + 1, height))
                y_1 = int(max(0, box[0] - patch_size))
                y_2 = int(min(box[2] + 1, width))
                close_patch_map[i, x_1:x_2, y_1:y_2] = False
            close_vulnerable = np.any(close_patch_map, axis=(-2, -1))

            # Over patch
            close_patch_map = np.ones((len(boxes), height, width), dtype=bool)
            for i, box in enumerate(boxes):
                x_1 = int(max(0, box[1] - patch_size))
                x_2 = int(min(box[3] + 1, height))
                y_1 = int(max(0, box[0] - patch_size))
                y_2 = int(min(box[2] + 1, width))
                close_patch_map[i, x_1:x_2, y_1:y_2] = True
            over_vulnerable = np.any(close_patch_map, axis=(-2, -1))

            cert = np.logical_and.reduce((far_vulnerable, close_vulnerable, over_vulnerable))
            certifications.append(cert)

        return certifications



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2023
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This module implements the abstract estimator `HuggingFaceClassifier` using the PyTorchClassifier as a backend
to interface with ART.
"""
import logging

from typing import List, Optional, Tuple, Union, Dict, Callable, Any, TYPE_CHECKING

import numpy as np
import six

from art.estimators.classification.pytorch import PyTorchClassifier

if TYPE_CHECKING:
    import torch
    import transformers
    from art.utils import CLIP_VALUES_TYPE, PREPROCESSING_TYPE
    from art.defences.preprocessor import Preprocessor
    from art.defences.postprocessor import Postprocessor
    from transformers.modeling_outputs import ImageClassifierOutput

logger = logging.getLogger(__name__)


class HuggingFaceClassifierPyTorch(PyTorchClassifier):
    """
    This class implements a classifier with the HuggingFace framework.
    """

    def __init__(
        self,
        model: "transformers.PreTrainedModel",
        loss: "torch.nn.modules.loss._Loss",
        input_shape: Tuple[int, ...],
        nb_classes: int,
        optimizer: Optional["torch.optim.Optimizer"] = None,
        use_amp: bool = False,
        opt_level: str = "O1",
        loss_scale: Optional[Union[float, str]] = "dynamic",
        channels_first: bool = True,
        clip_values: Optional["CLIP_VALUES_TYPE"] = None,
        preprocessing_defences: Union["Preprocessor", List["Preprocessor"], None] = None,
        postprocessing_defences: Union["Postprocessor", List["Postprocessor"], None] = None,
        preprocessing: "PREPROCESSING_TYPE" = (0.0, 1.0),
        processor: Optional[Callable] = None,
        device_type: str = "gpu",
    ):
        """
        Initialization of HuggingFaceClassifierPyTorch specifically for the PyTorch-based backend.

        :param model: Huggingface model model which returns outputs of type
                      ImageClassifierOutput from the transformers library.
                      Must have the logits attribute set as output.
        :param loss: The loss function for which to compute gradients for training. The target label must be raw
                categorical, i.e. not converted to one-hot encoding.
        :param input_shape: The shape of one input instance.
        :param optimizer: The optimizer used to train the classifier.
        :param use_amp: Whether to use the automatic mixed precision tool to enable mixed precision training or
                        gradient computation, e.g. with loss gradient computation. When set to True, this option is
                        only triggered if there are GPUs available.
        :param opt_level: Specify a pure or mixed precision optimization level. Used when use_amp is True. Accepted
                            values are `O0`, `O1`, `O2`, and `O3`.
        :param loss_scale: Loss scaling. Used when use_amp is True. If passed as a string, must be a string
                            representing a number, e.g., “1.0”, or the string “dynamic”.
        :param nb_classes: The number of classes of the model.
        :param optimizer: The optimizer used to train the classifier.
        :param channels_first: Set channels first or last. Normally should be set to True for HF models based on
                               a pytorch backend.
        :param clip_values: Tuple of the form `(min, max)` of floats or `np.ndarray` representing the minimum and
                maximum values allowed for features. If floats are provided, these will be used as the range of all
                features. If arrays are provided, each value will be considered the bound for a feature, thus
                the shape of clip values needs to match the total number of features.
        :param preprocessing_defences: Preprocessing defence(s) to be applied by the classifier.
        :param postprocessing_defences: Postprocessing defence(s) to be applied by the classifier.
        :param preprocessing: Tuple of the form `(subtrahend, divisor)` of floats or `np.ndarray` of values to be
                used for data preprocessing. The first value will be subtracted from the input. The input will then
                be divided by the second one.
        :param device_type: Type of device on which the classifier is run, either `gpu` or `cpu`.
        :param processor: Optional argument. Function which takes in a batch of data and performs
                          the preprocessing relevant to a given foundation model.
                          Must be differentiable for grandient based defences and attacks.
        """
        import torch

        self.processor = processor

        super().__init__(
            model=model,
            loss=loss,
            input_shape=input_shape,
            nb_classes=nb_classes,
            optimizer=optimizer,
            use_amp=use_amp,
            opt_level=opt_level,
            loss_scale=loss_scale,
            channels_first=channels_first,
            clip_values=clip_values,
            preprocessing_defences=preprocessing_defences,
            postprocessing_defences=postprocessing_defences,
            preprocessing=preprocessing,
            device_type=device_type,
        )

        import functools

        def prefix_function(function: Callable, postfunction: Callable) -> Callable[[Any, Any], torch.Tensor]:
            """
            Huggingface returns logit under outputs.logits. To make this compatible with ART we wrap the forward pass
            function of a HF model here, which automatically extracts the logits.

            :param function: The first function to run, in our case the forward pass of the model.
            :param postfunction: Second function to run, in this case simply extracts the logits.
            :return: model outputs.
            """

            @functools.wraps(function)
            def run(*args, **kwargs) -> torch.Tensor:
                outputs = function(*args, **kwargs)
                return postfunction(outputs)

            return run

        def get_logits(outputs: "ImageClassifierOutput") -> torch.Tensor:
            """
            Gets the logits attribute from ImageClassifierOutput

            :param outputs: outputs of type ImageClassifierOutput from a Huggingface
            :return: model's logit predictions.
            """
            if isinstance(outputs, torch.Tensor):
                return outputs
            return outputs.logits

        self.model.forward = prefix_function(self.model.forward, get_logits)  # type: ignore

    def _make_model_wrapper(self, model: "torch.nn.Module") -> "torch.nn.Module":
        # Try to import PyTorch and create an internal class that acts like a model wrapper extending torch.nn.Module
        import torch

        input_shape = self._input_shape
        input_for_hook = torch.rand(input_shape)
        # self.device may not match the device the raw model was passed into ART.
        # Check if the model is on cuda, if so set the hook input accordingly
        if next(model.parameters()).is_cuda:
            cuda_idx = torch.cuda.current_device()
            input_for_hook = input_for_hook.to(torch.device(f"cuda:{cuda_idx}"))

        input_for_hook = torch.unsqueeze(input_for_hook, dim=0)

        if self.processor is not None:
            input_for_hook = self.processor(input_for_hook)

        processor = self.processor
        try:
            # Define model wrapping class only if not defined before
            if not hasattr(self, "_model_wrapper"):

                class ModelWrapper(torch.nn.Module):
                    """
                    This is a wrapper for the input model.
                    """

                    def __init__(self, model: torch.nn.Module):
                        """
                        Initialization by storing the input model.

                        :param model: PyTorch model. The forward function of the model must return the logit output.
                        """
                        super().__init__()
                        self._model = model

                    # pylint: disable=W0221
                    # disable pylint because of API requirements for function
                    def forward(self, x):
                        """
                        This is where we get outputs from the input model.

                        :param x: Input data.
                        :type x: `torch.Tensor`
                        :return: a list of output layers, where the last 2 layers are logit and final outputs.
                        :rtype: `list`
                        """
                        # pylint: disable=W0212
                        # disable pylint because access to _model required

                        result = []

                        if isinstance(self._model, torch.nn.Module):
                            if processor is not None:
                                x = processor(x)
                            x = self._model.forward(x)
                            result.append(x)

                        else:  # pragma: no cover
                            raise TypeError("The input model must inherit from `nn.Module`.")

                        return result

                    @property
                    def get_layers(self) -> List[str]:
                        """
                        Return the hidden layers in the model, if applicable.

                        :return: The hidden layers in the model, input and output layers excluded.

                        .. warning:: `get_layers` tries to infer the internal structure of the model.
                                     This feature comes with no guarantees on the correctness of the result.
                                     The intended order of the layers tries to match their order in the model, but this
                                     is not guaranteed either.
                        """

                        result_dict = {}

                        modules = []

                        # pylint: disable=W0613
                        def forward_hook(input_module, hook_input, hook_output):
                            logger.info("input_module is %s with id %i", input_module, id(input_module))
                            modules.append(id(input_module))

                        handles = []

                        for name, module in self._model.named_modules():
                            logger.info(
                                "found %s with type %s and id %i and name %s with submods %i ",
                                module,
                                type(module),
                                id(module),
                                name,
                                len(list(module.named_modules())),
                            )

                            if name != "" and len(list(module.named_modules())) == 1:
                                handles.append(module.register_forward_hook(forward_hook))
                                result_dict[id(module)] = name

                        logger.info("mapping from id to name is %s", result_dict)

                        logger.info("------ Finished Registering Hooks------")
                        model(input_for_hook)  # hooks are fired sequentially from model input to the output

                        logger.info("------ Finished Fire Hooks------")

                        # Remove the hooks
                        for hook in handles:
                            hook.remove()

                        logger.info("new result is: ")
                        name_order = []
                        for module in modules:
                            name_order.append(result_dict[module])

                        logger.info(name_order)

                        return name_order

                # Set newly created class as private attribute
                self._model_wrapper = ModelWrapper  # type: ignore

            # Use model wrapping class to wrap the PyTorch model received as argument
            return self._model_wrapper(model)

        except ImportError:  # pragma: no cover
            raise ImportError("Could not find PyTorch (`torch`) installation.") from ImportError

    def get_activations(  # type: ignore
        self,
        x: Union[np.ndarray, "torch.Tensor"],
        layer: Optional[Union[int, str]] = None,
        batch_size: int = 128,
        framework: bool = False,
    ) -> Union[np.ndarray, "torch.Tensor"]:
        """
        Return the output of the specified layer for input `x`. `layer` is specified by layer index (between 0 and
        `nb_layers - 1`) or by name. The number of layers can be determined by counting the results returned by
        calling `layer_names`.

        :param x: Input for computing the activations.
        :param layer: Layer for computing the activations
        :param batch_size: Size of batches.
        :param framework: If true, return the intermediate tensor representation of the activation.
        :return: The output of `layer`, where the first dimension is the batch size corresponding to `x`.
        """
        import torch

        self._model.eval()

        # Apply defences
        if framework:
            no_grad = False
        else:
            no_grad = True
        x_preprocessed, _ = self._apply_preprocessing(x=x, y=None, fit=False, no_grad=no_grad)

        # Get index of the extracted layer
        if isinstance(layer, six.string_types):
            if layer not in self._layer_names:  # pragma: no cover
                raise ValueError(f"Layer name {layer} not supported")
            layer_index = self._layer_names.index(layer)

        elif isinstance(layer, int):
            layer_index = layer

        else:  # pragma: no cover
            raise TypeError("Layer must be of type str or int")

        def get_feature(name):
            # the hook signature
            def hook(model, input, output):  # pylint: disable=W0622,W0613
                # TODO: this is using the input, rather than the output, to circumvent the fact
                # TODO: that flatten is not a layer in pytorch, and the activation defence expects
                # TODO: a flattened input. A better option is to refactor the activation defence
                # TODO: to not crash if non 2D inputs are provided.
                self._features[name] = input

            return hook

        if not hasattr(self, "_features"):
            self._features: Dict[str, torch.Tensor] = {}
            # register forward hooks on the layers of choice
        handles = []

        lname = self._layer_names[layer_index]

        if layer not in self._features:
            for name, module in self.model.named_modules():
                if name == lname and len(list(module.named_modules())) == 1:
                    handles.append(module.register_forward_hook(get_feature(name)))

        if framework:
            if isinstance(x_preprocessed, torch.Tensor):
                self._model(x_preprocessed)
                return self._features[self._layer_names[layer_index]][0]
            input_tensor = torch.from_numpy(x_preprocessed)
            self._model(input_tensor.to(self._device))
            return self._features[self._layer_names[layer_index]][0]  # pylint: disable=W0212

        # Run prediction with batch processing
        results = []
        num_batch = int(np.ceil(len(x_preprocessed) / float(batch_size)))

        for m in range(num_batch):
            # Batch indexes
            begin, end = (
                m * batch_size,
                min((m + 1) * batch_size, x_preprocessed.shape[0]),
            )

            # Run prediction for the current batch
            self._model(torch.from_numpy(x_preprocessed[begin:end]).to(self._device))
            layer_output = self._features[self._layer_names[layer_index]]  # pylint: disable=W0212

            if isinstance(layer_output, tuple):
                results.append(layer_output[0].detach().cpu().numpy())
            else:
                results.append(layer_output.detach().cpu().numpy())

        results_array = np.concatenate(results)
        return results_array



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2019
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import pytest
import numpy as np

from art.attacks.evasion.fast_gradient import FastGradientMethod
from art.estimators.certification.randomized_smoothing import (
    NumpyRandomizedSmoothing,
    TensorFlowV2RandomizedSmoothing,
    PyTorchRandomizedSmoothing,
)
from art.utils import load_dataset, random_targets

from tests.utils import (
    get_image_classifier_pt,
    get_image_classifier_tf,
    get_image_classifier_kr,
    get_tabular_classifier_pt,
    ARTTestException,
)

logger = logging.getLogger(__name__)


@pytest.fixture()
def get_mnist_classifier(framework):
    def _get_classifier():
        if framework == "pytorch":
            import torch

            classifier = get_image_classifier_pt()
            optimizer = torch.optim.Adam(classifier.model.parameters(), lr=0.01)
            rs = PyTorchRandomizedSmoothing(
                model=classifier.model,
                loss=classifier._loss,
                optimizer=optimizer,
                input_shape=classifier.input_shape,
                nb_classes=classifier.nb_classes,
                channels_first=classifier.channels_first,
                clip_values=classifier.clip_values,
                sample_size=100,
                scale=0.01,
                alpha=0.001,
            )

        elif framework == "tensorflow2":
            classifier, _ = get_image_classifier_tf()
            rs = TensorFlowV2RandomizedSmoothing(
                model=classifier.model,
                nb_classes=classifier.nb_classes,
                input_shape=classifier.input_shape,
                loss_object=classifier.loss_object,
                optimizer=classifier.optimizer,
                train_step=classifier.train_step,
                channels_first=classifier.channels_first,
                clip_values=classifier.clip_values,
                preprocessing_defences=classifier.preprocessing_defences,
                postprocessing_defences=classifier.postprocessing_defences,
                preprocessing=classifier.preprocessing,
                sample_size=100,
                scale=0.01,
                alpha=0.001,
            )

        elif framework in ("keras", "kerastf"):
            classifier = get_image_classifier_kr()
            rs = NumpyRandomizedSmoothing(
                classifier=classifier,
                sample_size=100,
                scale=0.01,
                alpha=0.001,
            )

        else:
            classifier, rs = None, None

        return classifier, rs

    return _get_classifier


@pytest.fixture()
def get_iris_classifier(framework):
    def _get_classifier():
        if framework == "pytorch":
            classifier = get_tabular_classifier_pt()
            rs = PyTorchRandomizedSmoothing(
                model=classifier.model,
                loss=classifier._loss,
                input_shape=classifier.input_shape,
                nb_classes=classifier.nb_classes,
                channels_first=classifier.channels_first,
                clip_values=classifier.clip_values,
                sample_size=100,
                scale=0.01,
                alpha=0.001,
            )

        else:
            classifier, rs = None, None

        return classifier, rs

    return _get_classifier


@pytest.mark.only_with_platform("pytorch", "tensorflow2", "keras", "kerastf")
def test_randomized_smoothing_mnist_predict(art_warning, get_default_mnist_subset, get_mnist_classifier):
    (_, _), (x_test, y_test) = get_default_mnist_subset
    x_test, y_test = x_test[:10], y_test[:10]

    try:
        classifier, rs = get_mnist_classifier()
        y_test_base = classifier.predict(x=x_test)
        y_test_smooth = rs.predict(x=x_test)

        np.testing.assert_array_equal(y_test_smooth.shape, y_test_base.shape)
        np.testing.assert_array_almost_equal(np.sum(y_test_smooth, axis=1), np.ones(len(y_test)))
        np.testing.assert_array_almost_equal(np.argmax(y_test_smooth, axis=1), np.argmax(y_test_base, axis=1))

    except ARTTestException as e:
        art_warning(e)


@pytest.mark.only_with_platform("pytorch", "tensorflow2", "keras", "kerastf")
def test_randomized_smoothing_mnist_fit(art_warning, get_default_mnist_subset, get_mnist_classifier):
    (_, _), (x_test, y_test) = get_default_mnist_subset
    x_test, y_test = x_test[:10], y_test[:10]

    try:
        _, rs = get_mnist_classifier()
        rs.fit(x=x_test, y=y_test)

    except ARTTestException as e:
        art_warning(e)


@pytest.mark.only_with_platform("pytorch", "tensorflow2", "keras", "kerastf")
def test_randomized_smoothing_mnist_certify(art_warning, get_default_mnist_subset, get_mnist_classifier):
    (_, _), (x_test, y_test) = get_default_mnist_subset
    x_test, y_test = x_test[:10], y_test[:10]

    try:
        _, rs = get_mnist_classifier()
        pred, radius = rs.certify(x=x_test, n=250)

        np.testing.assert_array_equal(pred.shape, radius.shape)
        np.testing.assert_array_less(radius, 1)
        np.testing.assert_array_less(pred, y_test.shape[1])

    except ARTTestException as e:
        art_warning(e)


@pytest.mark.only_with_platform("pytorch", "tensorflow2", "keras", "kerastf")
def test_randomized_smoothing_mnist_loss_gradient(art_warning, get_default_mnist_subset, get_mnist_classifier):
    (_, _), (x_test, y_test) = get_default_mnist_subset
    x_test, y_test = x_test[:10], y_test[:10]

    try:
        _, rs = get_mnist_classifier()
        grad = rs.loss_gradient(x=x_test, y=y_test, sampling=True)

        np.testing.assert_array_equal(grad.shape, x_test.shape)

    except ARTTestException as e:
        art_warning(e)


@pytest.mark.only_with_platform("pytorch", "tensorflow2", "keras", "kerastf")
def test_randomized_smoothing_mnist_fgsm(art_warning, get_default_mnist_subset, get_mnist_classifier):
    (_, _), (x_test, y_test) = get_default_mnist_subset
    x_test, y_test = x_test[:10], y_test[:10]

    try:
        classifier, rs = get_mnist_classifier()
        fgsm = FastGradientMethod(estimator=classifier, targeted=True)
        params = {"y": random_targets(y_test, classifier.nb_classes)}
        x_test_adv = fgsm.generate(x_test, **params)

        fgsm_with_rs = FastGradientMethod(estimator=rs, targeted=True)
        x_test_adv_with_rs = fgsm_with_rs.generate(x_test, **params)

        np.testing.assert_array_equal(x_test_adv.shape, x_test_adv_with_rs.shape)
        np.testing.assert_array_less(np.abs(x_test_adv - x_test_adv_with_rs), 0.75)

    except ARTTestException as e:
        art_warning(e)


@pytest.mark.only_with_platform("pytorch")
def test_randomized_smoothing_iris_predict(art_warning, get_iris_classifier):
    (_, _), (x_test, y_test), _, _ = load_dataset("iris")

    try:
        _, rs = get_iris_classifier()
        y_test_smooth = rs.predict(x=x_test)

        np.testing.assert_array_equal(y_test_smooth.shape, y_test.shape)
        assert np.all(np.sum(y_test_smooth, axis=1) <= 1)

    except ARTTestException as e:
        art_warning(e)


@pytest.mark.only_with_platform("pytorch")
def test_randomized_smoothing_iris_certify(art_warning, get_iris_classifier):
    (_, _), (x_test, y_test), _, _ = load_dataset("iris")

    try:
        _, rs = get_iris_classifier()
        pred, radius = rs.certify(x=x_test, n=250)

        np.testing.assert_array_equal(pred.shape, radius.shape)
        np.testing.assert_array_less(radius, 1)
        np.testing.assert_array_less(pred, y_test.shape[1])

    except ARTTestException as e:
        art_warning(e)


@pytest.mark.only_with_platform("pytorch")
def test_randomized_smoothing_iris_fgsm(art_warning, get_iris_classifier):
    (_, _), (x_test, y_test), _, _ = load_dataset("iris")

    try:
        classifier, rs = get_iris_classifier()
        attack = FastGradientMethod(classifier, eps=0.1)
        x_test_adv = attack.generate(x_test)
        preds_smooth = np.argmax(rs.predict(x_test_adv), axis=1)

        assert not np.array_equal(x_test, x_test_adv)
        assert not np.array_equal(np.argmax(y_test, axis=1), preds_smooth)
        assert np.all(x_test_adv <= 1)
        assert np.all(x_test_adv >= 0)

    except ARTTestException as e:
        art_warning(e)



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2021
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This module implements Bullseye Polytope clean-label attacks on Neural Networks.
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import time
from typing import Optional, Tuple, Union, TYPE_CHECKING, List

import numpy as np
from tqdm.auto import trange

from art.attacks.attack import PoisoningAttackWhiteBox
from art.estimators import BaseEstimator, NeuralNetworkMixin
from art.estimators.classification.classifier import ClassifierMixin
from art.estimators.classification.pytorch import PyTorchClassifier

if TYPE_CHECKING:
    # pylint: disable=C0412
    import torch
    from art.utils import CLASSIFIER_NEURALNETWORK_TYPE

logger = logging.getLogger(__name__)


class BullseyePolytopeAttackPyTorch(PoisoningAttackWhiteBox):
    """
    Implementation of Bullseye Polytope Attack by Aghakhani, et. al. 2020.
    "Bullseye Polytope: A Scalable Clean-Label Poisoning Attack with Improved Transferability"

    This implementation is based on UCSB's original code here: https://github.com/ucsb-seclab/BullseyePoison

    | Paper link: https://arxiv.org/abs/2005.00191
    """

    attack_params = PoisoningAttackWhiteBox.attack_params + [
        "target",
        "feature_layer",
        "opt",
        "max_iter",
        "learning_rate",
        "momentum",
        "decay_iter",
        "decay_coeff",
        "epsilon",
        "norm",
        "dropout",
        "endtoend",
        "batch_size",
        "verbose",
    ]

    _estimator_requirements = (BaseEstimator, NeuralNetworkMixin, ClassifierMixin, PyTorchClassifier)

    def __init__(
        self,
        classifier: Union["CLASSIFIER_NEURALNETWORK_TYPE", List["CLASSIFIER_NEURALNETWORK_TYPE"]],
        target: np.ndarray,
        feature_layer: Union[Union[str, int], List[Union[str, int]]],
        opt: str = "adam",
        max_iter: int = 4000,
        learning_rate: float = 4e-2,
        momentum: float = 0.9,
        decay_iter: Union[int, List[int]] = 10000,
        decay_coeff: float = 0.5,
        epsilon: float = 0.1,
        dropout: float = 0.3,
        net_repeat: int = 1,
        endtoend: bool = True,
        batch_size: int = 128,
        verbose: bool = True,
    ):
        """
        Initialize an Feature Collision Clean-Label poisoning attack

        :param classifier: The proxy classifiers used for the attack. Can be a single classifier or list of classifiers
                           with varying architectures.
        :param target: The target input(s) of shape (N, W, H, C) to misclassify at test time. Multiple targets will be
                       averaged.
        :param feature_layer: The name(s) of the feature representation layer(s).
        :param opt: The optimizer to use for the attack. Can be 'adam' or 'sgd'
        :param max_iter: The maximum number of iterations for the attack.
        :param learning_rate: The learning rate of clean-label attack optimization.
        :param momentum: The momentum of clean-label attack optimization.
        :param decay_iter: Which iterations to decay the learning rate.
                           Can be a integer (every N iterations) or list of integers [0, 500, 1500]
        :param decay_coeff: The decay coefficient of the learning rate.
        :param epsilon: The perturbation budget
        :param dropout: Dropout to apply while training
        :param net_repeat: The number of times to repeat prediction on each network
        :param endtoend: True for end-to-end training. False for transfer learning.
        :param batch_size: Batch size.
        :param verbose: Show progress bars.
        """
        self.subsistute_networks: List["CLASSIFIER_NEURALNETWORK_TYPE"] = (
            [classifier] if not isinstance(classifier, list) else classifier
        )

        super().__init__(classifier=self.subsistute_networks[0])  # type: ignore
        self.target = target
        self.opt = opt
        self.momentum = momentum
        self.decay_iter = decay_iter
        self.epsilon = epsilon
        self.dropout = dropout
        self.net_repeat = net_repeat
        self.endtoend = endtoend
        self.feature_layer = feature_layer
        self.learning_rate = learning_rate
        self.decay_coeff = decay_coeff
        self.max_iter = max_iter
        self.batch_size = batch_size
        self.verbose = verbose
        self._check_params()

    def poison(self, x: np.ndarray, y: Optional[np.ndarray] = None, **kwargs) -> Tuple[np.ndarray, np.ndarray]:
        """
        Iteratively finds optimal attack points starting at values at x

        :param x: The base images to begin the poison process.
        :param y: Target label
        :return: An tuple holding the (poisoning examples, poisoning labels).
        """
        import torch

        class PoisonBatch(torch.nn.Module):
            """
            Implementing this to work with PyTorch optimizers.
            """

            def __init__(self, base_list):
                super().__init__()
                base_batch = torch.stack(base_list, 0)
                self.poison = torch.nn.Parameter(base_batch.clone())

            def forward(self):
                """Forward method."""
                return self.poison

        base_tensor_list = [torch.from_numpy(sample).to(self.estimator.device) for sample in x]
        poison_batch = PoisonBatch([torch.from_numpy(np.copy(sample)).to(self.estimator.device) for sample in x])
        opt_method = self.opt.lower()

        if opt_method == "sgd":  # pragma: no cover
            logger.info("Using SGD to craft poison samples")
            optimizer = torch.optim.SGD(poison_batch.parameters(), lr=self.learning_rate, momentum=self.momentum)
        elif opt_method == "adam":
            logger.info("Using Adam to craft poison samples")
            optimizer = torch.optim.Adam(  # type: ignore
                poison_batch.parameters(), lr=self.learning_rate, betas=(self.momentum, 0.999)
            )

        base_tensor_batch = torch.stack(base_tensor_list, 0)
        base_range01_batch = base_tensor_batch

        # Because we have turned on DP for the substitute networks,
        # the target image's feature becomes random.
        # We can try enforcing the convex polytope in one of the multiple realizations of the feature,
        # but empirically one realization is enough.
        target_feat_list = []
        # Coefficients for the convex combination.
        # Initializing from the coefficients of last step gives faster convergence.
        s_init_coeff_list = []
        n_poisons = len(x)
        s_coeff: Union["torch.Tensor", List["torch.Tensor"]]
        for _, net in enumerate(self.subsistute_networks):
            # End to end training
            if self.endtoend:
                if isinstance(self.feature_layer, list):
                    block_feats = []
                    for layer in self.feature_layer:
                        activations = net.get_activations(x, layer=layer, batch_size=self.batch_size, framework=True)
                        if activations is not None:
                            block_feats += [torch.stack([feat.detach() for feat in activations], 0)]
                        else:  # pragma: no cover
                            raise ValueError("Activations are None.")
                else:
                    layer_2: Union[int, str] = self.feature_layer
                    activations = net.get_activations(x, layer=layer_2, batch_size=self.batch_size, framework=True)
                    if activations is not None:
                        block_feats = [feat.detach() for feat in activations]
                    else:  # pragma: no cover
                        raise ValueError("Activations are None.")
                target_feat_list.append(block_feats)
                s_coeff = [
                    torch.ones(n_poisons, 1).to(self.estimator.device) / n_poisons for _ in range(len(block_feats))
                ]
            else:  # pragma: no cover
                if isinstance(self.feature_layer, list):
                    raise NotImplementedError
                layer_3: Union[int, str] = self.feature_layer
                activations = net.get_activations(x, layer=layer_3, batch_size=self.batch_size, framework=True)
                if activations is not None:
                    target_feat_list.append(activations.detach())  # type: ignore
                else:
                    raise ValueError("Activations are None.")
                s_coeff = torch.ones(n_poisons, 1).to(self.estimator.device) / n_poisons

            s_init_coeff_list.append(s_coeff)

        for ite in trange(self.max_iter):
            if ite % self.decay_iter == 0 and ite != 0:
                for param_group in optimizer.param_groups:
                    param_group["lr"] *= self.decay_coeff
                print(f"{time.strftime('%Y-%m-%d %H:%M:%S')} Iteration {ite}, Adjusted lr to {self.learning_rate:.2e}")

            poison_batch.zero_grad()
            total_loss = loss_from_center(
                self.subsistute_networks,
                target_feat_list,
                poison_batch,
                self.net_repeat,
                self.endtoend,
                self.feature_layer,
            )
            total_loss.backward()
            optimizer.step()

            # clip the perturbations into the range
            perturb_range01 = torch.clamp((poison_batch.poison.data - base_tensor_batch), -self.epsilon, self.epsilon)
            perturbed_range01 = torch.clamp(
                base_range01_batch.data + perturb_range01.data,
                self.estimator.clip_values[0],
                self.estimator.clip_values[1],
            )
            poison_batch.poison.data = perturbed_range01

        if y is None:  # pragma: no cover
            raise ValueError("You must pass in the target label as y")

        return get_poison_tuples(poison_batch, y)

    def _check_params(self) -> None:
        if self.learning_rate <= 0:
            raise ValueError("Learning rate must be strictly positive")

        if self.max_iter < 1:
            raise ValueError("Value of max_iter at least 1")

        if not isinstance(self.feature_layer, (str, int, list)):
            raise TypeError("Feature layer should be a string or int or list of string or int")

        if self.opt.lower() not in ["adam", "sgd"]:
            raise ValueError("Optimizer must be 'adam' or 'sgd'")

        if not 0 <= self.momentum <= 1:
            raise ValueError("Momentum must be between 0 and 1")

        if isinstance(self.decay_iter, int) and self.decay_iter < 0:
            raise ValueError("decay_iter must be at least 0")

        if isinstance(self.decay_iter, list) and not all(
            (isinstance(decay_iter, int) and decay_iter > 0 for decay_iter in self.decay_iter)
        ):
            raise ValueError("decay_iter is not a list of positive integers")

        if self.epsilon <= 0:
            raise ValueError("epsilon must be at least 0")

        if not 0 <= self.dropout <= 1:
            raise ValueError("dropout must be between 0 and 1")

        if self.net_repeat < 1:
            raise ValueError("net_repeat must be at least 1")

        if isinstance(self.feature_layer, list):
            for layer in self.feature_layer:
                if isinstance(layer, int):
                    if not 0 <= layer < len(self.estimator.layer_names):
                        raise ValueError("feature_layer is not list of positive integers")
                elif not isinstance(layer, str):
                    raise ValueError("feature_layer is not list of strings")

        if isinstance(self.feature_layer, int):
            if not 0 <= self.feature_layer < len(self.estimator.layer_names):
                raise ValueError("feature_layer is not positive integer")

        if not 0 <= self.decay_coeff <= 1:
            raise ValueError("Decay coefficient must be between zero and one")

        if not isinstance(self.batch_size, int) or self.batch_size <= 0:
            raise ValueError("batch_size must be a positive integer")


def get_poison_tuples(poison_batch, poison_label):
    """
    Includes the labels
    """
    poison = [
        poison_batch.poison.data[num_p].unsqueeze(0).detach().cpu().numpy()
        for num_p in range(poison_batch.poison.size(0))
    ]
    return np.vstack(poison), poison_label


def loss_from_center(
    subs_net_list, target_feat_list, poison_batch, net_repeat, end2end, feature_layer
) -> "torch.Tensor":
    """
    Calculate loss from center.
    """
    import torch

    if end2end:
        loss = torch.tensor(0.0)
        for net, center_feats in zip(subs_net_list, target_feat_list):
            poisons_feats: Union[List[float], "torch.Tensor", np.ndarray]
            if net_repeat > 1:
                poisons_feats_repeats = [
                    net.get_activations(poison_batch(), layer=feature_layer, framework=True) for _ in range(net_repeat)
                ]
                block_num = len(poisons_feats_repeats[0])
                poisons_feats = []
                for block_idx in range(block_num):
                    poisons_feats.append(
                        sum([poisons_feat_r[block_idx] for poisons_feat_r in poisons_feats_repeats]) / net_repeat
                    )
            elif net_repeat == 1:
                if isinstance(feature_layer, list):
                    poisons_feats = [
                        torch.flatten(
                            net.get_activations(poison_batch(), layer=layer, framework=True), 0  # type: ignore
                        )
                        for layer in feature_layer
                    ]
                else:  # pragma: no cover
                    poisons_feats = net.get_activations(poison_batch(), layer=feature_layer, framework=True)
            else:  # pragma: no cover
                assert False, f"net_repeat set to {net_repeat}"

            net_loss = torch.tensor(0.0)
            for pfeat, cfeat in zip(poisons_feats, center_feats):
                diff = torch.mean(pfeat, dim=0) - cfeat
                diff_norm = torch.norm(diff, dim=0)
                cfeat_norm = torch.norm(cfeat, dim=0)
                diff_norm = diff_norm / cfeat_norm
                net_loss += torch.mean(diff_norm)
            loss += net_loss / len(center_feats)
        loss = loss / len(subs_net_list)

    else:  # pragma: no cover
        loss = torch.tensor(0.0)
        for net, center in zip(subs_net_list, target_feat_list):
            poisons_list = [
                net.get_activations(poison_batch(), layer=feature_layer, framework=True) for _ in range(net_repeat)
            ]
            poisons = torch.tensor(sum(poisons_list) / len(poisons_list))

            diff_2 = torch.mean(poisons, dim=0) - center
            diff_norm = torch.norm(diff_2, dim=1) / torch.norm(center, dim=1)
            loss += torch.mean(diff_norm)

        loss = loss / len(subs_net_list)

    return loss



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2022
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# MIT License
#
# Copyright (c) 2022 University of Michigan and University of Wisconsin-Madison
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This module implements the black-box (hard-label) GRAPHITE attack `GRAPHITEBlackbox`. This is a physical black-box
attack that only requires class predictions.

| Paper link: https://arxiv.org/abs/2002.07088
| Original github link: https://github.com/ryan-feng/GRAPHITE
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import logging
from typing import Optional, Tuple, Union, TYPE_CHECKING, List

import random
import numpy as np

from art.config import ART_NUMPY_DTYPE
from art.attacks.attack import EvasionAttack
from art.estimators.estimator import BaseEstimator
from art.estimators.classification import ClassifierMixin
from art.utils import compute_success, to_categorical, check_and_transform_label_format
from art.attacks.evasion.graphite.utils import (
    convert_to_network,
    get_transform_params,
    add_noise,
    get_transformed_images,
    run_predictions,
    score_fn,
)

if TYPE_CHECKING:
    from art.utils import CLASSIFIER_NEURALNETWORK_TYPE

logger = logging.getLogger(__name__)


class GRAPHITEBlackbox(EvasionAttack):
    """
    Implementation of the hard-label GRAPHITE attack from Feng et al. (2022). This is a physical, black-box attack
    that only requires final class prediction and generates robust physical perturbations that can be applied as
    stickers.

    | Paper link: https://arxiv.org/abs/2002.07088
    | Original github link: https://github.com/ryan-feng/GRAPHITE
    """

    attack_params = EvasionAttack.attack_params + [
        "noise_size",
        "net_size",
        "heat_patch_size",
        "heat_patch_stride",
        "heatmap_mode",
        "tr_lo",
        "tr_hi",
        "num_xforms_mask",
        "max_mask_size",
        "rotation_range",
        "beta",
        "eta",
        "num_xforms_boost",
        "num_boost_queries",
        "dist_range",
        "gamma_range",
        "crop_percent_range",
        "off_x_range",
        "off_y_range",
        "blur_kernels",
        "batch_size",
    ]
    _estimator_requirements = (BaseEstimator, ClassifierMixin)

    def __init__(
        self,
        classifier: "CLASSIFIER_NEURALNETWORK_TYPE",
        noise_size: Tuple[int, int],
        net_size: Tuple[int, int],
        heat_patch_size: Tuple[int, int] = (4, 4),
        heat_patch_stride: Tuple[int, int] = (1, 1),
        heatmap_mode: str = "Target",
        tr_lo: float = 0.65,
        tr_hi: float = 0.85,
        num_xforms_mask: int = 100,
        max_mask_size: int = -1,
        beta: float = 1.0,
        eta: float = 500,
        num_xforms_boost: int = 100,
        num_boost_queries: int = 20000,
        rotation_range: Tuple[float, float] = (-30.0, 30.0),
        dist_range: Tuple[float, float] = (0.0, 0.0),
        gamma_range: Tuple[float, float] = (1.0, 2.0),
        crop_percent_range: Tuple[float, float] = (-0.03125, 0.03125),
        off_x_range: Tuple[float, float] = (-0.03125, 0.03125),
        off_y_range: Tuple[float, float] = (-0.03125, 0.03125),
        blur_kernels: Union[Tuple[int, int], List[int]] = (0, 3),
        batch_size: int = 64,
    ) -> None:
        """
        Create a GRAPHITEBlackbox attack instance.

        :param classifier: A trained classifier.
        :param noise_size: The resolution to generate perturbations in (w, h).
        :param net_size: The resolution to resize images to before feeding to the model in (w, h).
        :param heat_patch_size: The size of the heatmap patches in (w, h).
        :param heat_patch_stride: The stride of the heatmap patching in (w, h).
        :param heatmap_mode: The mode of heatmap in ['Target', 'Random'].
        :param tr_lo: tr_lo, threshold for fine-grained reduction.
        :param tr_hi: tr_hi, threshold for coarse-grained reduction.
        :param num_xforms_mask: The number of transforms to use in mask generation.
        :param max_mask_size: Optionally specify that you just want to optimize until a mask size of <= max_mask_size.
        :param beta: The parameter beta for RGF optimization in boosting.
        :param eta: The step size for RGF optimization in boosting.
        :param num_xforms_boost: The number of transforms to use in boosting.
        :param num_boost_queries: The number of queries to use in boosting.
        :param rotation_range: The range of the rotation in the perspective transform.
        :param dist_range: The range of the distance (in ft) to be added to the focal length in perspective transform.
        :param gamma_range: The range of the gamma in the gamma transform.
        :param crop_percent_range: The range of the crop percent in the perspective transform.
        :param off_x_range: The range of the x offset (percent) in the perspective transform.
        :param off_y_range: The range of the y offset (percent) in the perspective transform.
        :param blur_kernels: The kernels to blur with.
        :param batch_size: The size of the batch used by the estimator during inference.
        """
        super().__init__(estimator=classifier)
        self.noise_size = noise_size
        self.net_size = net_size
        self.heat_patch_size = heat_patch_size
        self.heat_patch_stride = heat_patch_stride
        self.heatmap_mode = heatmap_mode
        self.batch_size = batch_size
        self.tr_lo = tr_lo
        self.tr_hi = tr_hi
        self.num_xforms_mask = num_xforms_mask
        self.max_mask_size = max_mask_size
        self.beta = beta
        self.eta = eta
        self.num_xforms_boost = num_xforms_boost
        self.num_boost_queries = num_boost_queries
        self.rotation_range = rotation_range
        self.dist_range = dist_range
        self.gamma_range = gamma_range
        self.crop_percent_range = crop_percent_range
        self.off_x_range = off_x_range
        self.off_y_range = off_y_range
        self.blur_kernels = blur_kernels

        self._check_params()

    def generate(self, x: np.ndarray, y: Optional[np.ndarray] = None, **kwargs) -> np.ndarray:
        """
        Generate adversarial samples and return them in an array.

        :param x: An array with the original inputs to be attacked.
        :param y: Target values (class labels) one-hot-encoded of shape `(nb_samples, nb_classes)` or indices of shape
                  (nb_samples,).
        :param mask: An array with a mask broadcastable to input `x` defining where to apply adversarial perturbations.
                     Shape needs to be broadcastable to the shape of x and can also be of the same shape as `x`. Any
                     features for which the mask is zero will not be adversarially perturbed.
        :param x_tar: Initial array to act as the example target image.
        :param pts: Optional points to consider when cropping the perspective transform. An array of points in
                    [x, y, scale] with shape [num points, 3, 1].
        :param obj_width: The estimated object width (inches) for perspective transform. 30 by default.
        :param focal: The estimated focal length (ft) for perspective transform. 3 by default.
        :return: An array holding the adversarial examples.
        """
        mask = kwargs.get("mask")
        x_tar = kwargs.get("x_tar")
        obj_width = kwargs.get("obj_width") if "obj_width" in kwargs else 30
        focal = kwargs.get("focal") if "focal" in kwargs else 3
        pts = kwargs.get("pts") if "pts" in kwargs else None

        if y is None:
            raise ValueError("Target labels `y` need to be provided.")

        if x_tar is None:
            raise ValueError("Target image example `x_tar` needs to be provided.")

        y = check_and_transform_label_format(y, self.estimator.nb_classes)

        if self.estimator.nb_classes == 2 and y.shape[1] == 1:  # pragma: no cover
            raise ValueError(
                "This attack has not yet been tested for binary classification with a single output classifier."
            )

        if not isinstance(obj_width, int) and not isinstance(obj_width, float):
            raise ValueError("obj_width must be int or float")
        obj_width = float(obj_width)

        if not isinstance(focal, int) and not isinstance(focal, float):
            raise ValueError("focal must be int or float")
        focal = float(focal)

        # Check the mask
        if mask is not None:
            if len(mask.shape) == len(x.shape):
                mask = mask.astype(ART_NUMPY_DTYPE)
            else:
                mask = np.array([mask.astype(ART_NUMPY_DTYPE)] * x.shape[0])
        else:
            mask = np.array([None] * x.shape[0])

        # Get clip_min and clip_max from the classifier or infer them from data
        if self.estimator.clip_values is not None:
            clip_min, clip_max = self.estimator.clip_values
        else:
            clip_min, clip_max = np.min(x), np.max(x)

        # target initialization image
        x_tar = kwargs.get("x_tar")

        # Some initial setups
        x_adv = x.astype(ART_NUMPY_DTYPE)

        if self.estimator.channels_first:
            x = np.transpose(x, (0, 2, 3, 1))
            x_adv = np.transpose(x_adv, (0, 2, 3, 1))
            x_tar = np.transpose(x_tar, (0, 2, 3, 1))
            if len(mask.shape) == 4:
                mask = np.transpose(mask, (0, 2, 3, 1))

        y = np.argmax(y, axis=1)

        # Generate the adversarial samples
        for i in range(x_adv.shape[0]):
            x_adv[i] = self._perturb(
                x=x_adv[i],
                y=y[i],  # type: ignore
                x_tar=x_tar[i],
                obj_width=obj_width,
                focal=focal,
                clip_min=clip_min,
                clip_max=clip_max,
                mask=mask[i],
                pts=pts,
            )

        y = to_categorical(y, self.estimator.nb_classes)  # type: ignore

        # COMPUTE SUCCESS RATE
        x_copy = np.zeros((x.shape[0], self.noise_size[1], self.noise_size[0], x.shape[3]))
        x_adv_copy = np.zeros((x_adv.shape[0], self.noise_size[1], self.noise_size[0], x_adv.shape[3]))
        for i in range(x_copy.shape[0]):
            x_copy[i] = convert_to_network(x[i], self.net_size, clip_min, clip_max)
            x_adv_copy[i] = convert_to_network(x_adv[i], self.net_size, clip_min, clip_max)

        if self.estimator.channels_first:
            x_copy = np.transpose(x_copy, (0, 3, 1, 2))
            x_adv_copy = np.transpose(x_adv_copy, (0, 3, 1, 2))

        logger.info(
            "Success rate of GRAPHITE hard-label attack: %.2f%%",
            100
            * compute_success(
                self.estimator,
                x_copy.astype(np.float32),
                y,
                x_adv_copy.astype(np.float32),
                self.targeted,
                batch_size=self.batch_size,
            ),
        )

        if self.estimator.channels_first:
            x_adv = np.transpose(x_adv, (0, 3, 1, 2))

        return x_adv

    def _perturb(
        self,
        x: np.ndarray,
        y: int,
        x_tar: np.ndarray,
        obj_width: float,
        focal: float,
        clip_min: float,
        clip_max: float,
        mask: Optional[np.ndarray] = None,
        pts: Optional[np.ndarray] = None,
    ) -> np.ndarray:
        """
        Internal attack function for one example.

        :param x: An array with one original input to be attacked.
        :param y: The target label.
        :param x_tar: Initial array to act as an example target image.
        :param obj_width: Estimated width of object in inches for perspective transform.
        :param focal: Estimated focal length in ft for perspective transform.
        :param clip_min: Minimum value of an example.
        :param clip_max: Maximum value of an example.
        :param mask: An array with a mask to be applied to the adversarial perturbations. Shape needs to be
                     broadcastable to the shape of x. Any features for which the mask is zero will not be adversarially
                     perturbed.
        :param pts: Optional. A set of points that will set the crop size in the perspective transform.
        :return: An adversarial example.
        """
        import cv2

        x = (x.copy() - clip_min) / (clip_max - clip_min)
        x_tar = (x_tar.copy() - clip_min) / (clip_max - clip_min)

        if mask is None:
            mask_array = np.ones((self.noise_size[1], self.noise_size[0], x.shape[2]))
        else:
            mask_array = mask
        mask_array = mask_array / np.max(mask_array)

        x_copy = x.copy()
        x_tar_copy = x_tar.copy()
        mask_copy = mask_array.copy()
        x_noise = cv2.resize(x_copy, self.noise_size)
        x_tar_noise = cv2.resize(x_tar_copy, self.noise_size)
        mask_noise = cv2.resize(mask_copy, self.noise_size).astype(float)
        mask_noise = np.where(mask_noise > 0.5, 1.0, 0.0)

        if len(x_noise.shape) < 3:
            x_noise = x_noise[:, :, np.newaxis]
        if len(x_tar_noise.shape) < 3:
            x_tar_noise = x_tar_noise[:, :, np.newaxis]
        if len(mask_noise.shape) < 3:
            mask_noise = mask_noise[:, :, np.newaxis]

        mask_out = self._generate_mask(
            x_copy, x_noise, x_tar_noise, mask_noise, y, obj_width, focal, clip_min, clip_max, pts
        )

        adversarial = self._boost(x_copy, x_noise, x_tar_noise, mask_out, y, obj_width, focal, clip_min, clip_max, pts)

        adversarial = np.clip(adversarial.copy() * (clip_max - clip_min) + clip_min, clip_min, clip_max)
        return adversarial

    def _generate_mask(
        self,
        x: np.ndarray,
        x_noise: np.ndarray,
        x_tar_noise: np.ndarray,
        mask: np.ndarray,
        y: int,
        obj_width: float,
        focal: float,
        clip_min: float,
        clip_max: float,
        pts: Optional[np.ndarray] = None,
    ) -> np.ndarray:
        """
        Function to generate a mask.

        :param x: An array with one original input to be attacked.
        :param x_noise: x in the resolution of the noise size.
        :param x_tar_noise: x_tar in the resolution of the noise size.
        :param mask: An array with a mask to be applied to the adversarial perturbations. Shape needs to be
                     broadcastable to the shape of x. Any features for which the mask is zero will not be adversarially
                     perturbed.
        :param y: The target label.
        :param obj_width: Estimated width of object in inches for perspective transform.
        :param focal: Estimated focal length in ft for perspective transform.
        :param clip_min: Minimum value of an example.
        :param clip_max: Maximum value of an example.
        :param pts: Optional. A set of points that will set the crop size in the perspective transform.
        :return: A mask.
        """

        # Stage 1: HEATMAP: Collect all the valid patches and order by computed heatmap
        xforms = get_transform_params(
            self.num_xforms_mask,
            self.rotation_range,
            self.dist_range,
            self.gamma_range,
            self.crop_percent_range,
            self.off_x_range,
            self.off_y_range,
            self.blur_kernels,
            obj_width,
            focal,
        )

        object_size = mask.sum() / mask.shape[-1]

        patch = np.ones((self.heat_patch_size[1], self.heat_patch_size[0], x.shape[-1]))
        patches = []
        indices = []

        # collect all valid patches
        for i in range(0, mask.shape[0] - self.heat_patch_size[1] + 1, self.heat_patch_stride[1]):
            for j in range(0, mask.shape[1] - self.heat_patch_size[0] + 1, self.heat_patch_stride[0]):
                new_mask = np.zeros(mask.shape)
                new_mask[
                    i : min(i + self.heat_patch_size[1], mask.shape[0]),
                    j : min(j + self.heat_patch_size[0], mask.shape[1]),
                ] = patch
                new_mask = new_mask * mask
                if np.sum(new_mask) > 0:
                    patches.append(new_mask)
                    indices.append((i, j))

        # compute heatmap and order
        if self.heatmap_mode == "Random":
            tr_scores = [random.random() for i in range(len(patches))]
        else:  # Target mode

            tr_scores = self._get_heatmap(
                x,
                x_noise,
                x_tar_noise,
                mask,
                y,
                patches,
                xforms,
                clip_min,
                clip_max,
                pts,
            )

        tr_scores_np = np.asarray(tr_scores)
        order = tr_scores_np.argsort()
        patches = [patches[ind] for ind in order]
        indices = [indices[ind] for ind in order]

        # STAGE 2: Coarse Reduction
        best_mask, patches, indices = self._get_coarse_reduced_mask(
            x, x_noise, x_tar_noise, y, mask, patches, indices, xforms, clip_min, clip_max, pts
        )

        # STAGE 3: Fine Reduction
        if self.max_mask_size > 0:
            lbd = 5
            while best_mask.sum() / mask.shape[-1] > self.max_mask_size:
                patches_copy = list(patches)
                best_mask = self._get_fine_reduced_mask(
                    x,
                    x_noise,
                    x_tar_noise,
                    y,
                    best_mask,
                    patches_copy,
                    xforms,
                    object_size,
                    clip_min,
                    clip_max,
                    lbd,
                    pts,
                )
                lbd += 5
        else:
            best_mask = self._get_fine_reduced_mask(
                x,
                x_noise,
                x_tar_noise,
                y,
                best_mask,
                patches,
                xforms,
                object_size,
                clip_min,
                clip_max,
                pts=pts,
            )

        return best_mask

    def _get_heatmap(
        self,
        x: np.ndarray,
        x_noise: np.ndarray,
        x_tar_noise: np.ndarray,
        mask: np.ndarray,
        y: int,
        patches: List[np.ndarray],
        xforms: List,
        clip_min: float,
        clip_max: float,
        pts: Optional[np.ndarray] = None,
    ) -> List[float]:
        """
        Function to generate a heatmap.

        :param x: An array with one original input to be attacked.
        :param x_noise: x in the resolution of the noise size.
        :param x_tar_noise: x_tar in the resolution of the noise size.
        :param mask: An array with a mask to be applied to the adversarial perturbations. Shape needs to be
                     broadcastable to the shape of x. Any features for which the mask is zero will not be adversarially
                     perturbed.
        :param y: The target label.
        :param patches: list of patches from heatmap.
        :param xforms: list of transform params.
        :param clip_min: Minimum value of an example.
        :param clip_max: Maximum value of an example.
        :param pts: Optional. A set of points that will set the crop size in the perspective transform.
        :return: List of transform-robustness scores for the list of patches.
        """
        tr_scores = []

        # iterate over patches and compute transform_robustness without each individual patch
        for patch in patches:
            next_mask = mask * (np.ones(mask.shape) - patch)
            theta = (x_tar_noise - x_noise) * next_mask
            xform_imgs = get_transformed_images(
                x, next_mask, xforms, 1.0, theta, self.net_size, clip_min, clip_max, pts
            )
            success_rate = run_predictions(self.estimator, xform_imgs, y, self.batch_size, False)
            tr_scores.append(success_rate)

        return tr_scores

    def _evaluate_transform_robustness_at_pivot(
        self,
        x: np.ndarray,
        x_noise: np.ndarray,
        x_tar_noise: np.ndarray,
        y: int,
        mask: np.ndarray,
        patches: List[np.ndarray],
        xforms: List,
        clip_min: float,
        clip_max: float,
        pivot: int,
        pts: Optional[np.ndarray] = None,
    ) -> Tuple[float, np.ndarray]:
        """
        Function as a binary search plug-in that evaluates the transform-robustness at the specified pivot.

        :param x: An array with one original input to be attacked.
        :param x_noise: x in the resolution of the noise size.
        :param x_tar_noise: x_tar in the resolution of the noise size.
        :param y: The target label.
        :param mask: An array with a mask to be applied to the adversarial perturbations. Shape needs to be
                     broadcastable to the shape of x. Any features for which the mask is zero will not be adversarially
                     perturbed.
        :param patches: list of patches from heatmap.
        :param xforms: list of transform params.
        :param clip_min: Minimum value of an example.
        :param clip_max: Maximum value of an example.
        :param pivot: Pivot point to evaluate transform-robustness at.
        :param pts: Optional. A set of points that will set the crop size in the perspective transform.
        :return: transform-robustness and mask.
        """

        best_mask = np.zeros(mask.shape)
        ordering = patches[:pivot]
        for next_patch in ordering:
            next_mask = best_mask + (np.zeros(best_mask.shape) + next_patch)
            next_mask = np.where(next_mask > 0, 1.0, 0.0)
            best_mask = next_mask

        theta = (x_tar_noise - x_noise) * best_mask
        xform_imgs = get_transformed_images(x, best_mask, xforms, 1.0, theta, self.net_size, clip_min, clip_max, pts)
        success_rate = run_predictions(self.estimator, xform_imgs, y, self.batch_size, False)

        return (success_rate, best_mask)

    def _get_coarse_reduced_mask(
        self,
        x: np.ndarray,
        x_noise: np.ndarray,
        x_tar_noise: np.ndarray,
        y: int,
        mask: np.ndarray,
        patches: List[np.ndarray],
        indices: List[Tuple[int, int]],
        xforms: List,
        clip_min: float,
        clip_max: float,
        pts: Optional[np.ndarray] = None,
    ) -> Tuple[np.ndarray, List[np.ndarray], List[Tuple[int, int]]]:
        """
        Function to coarsely reduce mask.

        :param x: An array with one original input to be attacked.
        :param x_noise: x in the resolution of the noise size.
        :param x_tar_noise: x_tar in the resolution of the noise size.
        :param y: The target label.
        :param mask: An array with a mask to be applied to the adversarial perturbations. Shape needs to be
                     broadcastable to the shape of x. Any features for which the mask is zero will not be adversarially
                     perturbed.
        :param patches: list of patches from heatmap.
        :param indices: list of indices for the heatmap patches.
        :param xforms: list of transform params.
        :param clip_min: Minimum value of an example.
        :param clip_max: Maximum value of an example.
        :param pts: Optional. A set of points that will set the crop size in the perspective transform.
        :return: mask, adjusted list of patches, adjusted list of indices
        """
        # binary search leftmost pivot value for which tr exceeeds specificed threshold if one exists
        num_patches = len(patches)
        if num_patches == 1:
            pivot = 0

        else:
            low = 0
            high = num_patches - 1
            while low <= high:
                mid = low + (high - low) // 2
                score, _ = self._evaluate_transform_robustness_at_pivot(
                    x, x_noise, x_tar_noise, y, mask, patches, xforms, clip_min, clip_max, mid, pts
                )
                if score >= self.tr_hi:
                    if mid > 0:
                        high = mid - 1
                        continue
                    break

                # score < threshold:
                low = mid + 1

            pivot = mid

        _, best_mask = self._evaluate_transform_robustness_at_pivot(
            x, x_noise, x_tar_noise, y, mask, patches, xforms, clip_min, clip_max, pivot, pts
        )

        patches = patches[:]  # reduce will examine all patches
        indices = indices[:]

        return best_mask, patches, indices

    def _get_fine_reduced_mask(
        self,
        x: np.ndarray,
        x_noise: np.ndarray,
        x_tar_noise: np.ndarray,
        y: int,
        mask: np.ndarray,
        patches: List[np.ndarray],
        xforms: List,
        object_size: float,
        clip_min: float,
        clip_max: float,
        lbd: float = 5,
        pts: Optional[np.ndarray] = None,
    ) -> np.ndarray:
        """
        Function to finely reduce mask.

        :param x: An array with one original input to be attacked.
        :param x_noise: x in the resolution of the noise size.
        :param x_tar_noise: x_tar in the resolution of the noise size.
        :param y: The target label.
        :param mask: An array with a mask to be applied to the adversarial perturbations. Shape needs to be
                     broadcastable to the shape of x. Any features for which the mask is zero will not be adversarially
                     perturbed.
        :param patches: list of patches from heatmap.
        :param xforms: list of transform params.
        :param obj_size: Estimated width of object in inches for perspective transform.
        :param clip_min: Minimum value of an example.
        :param clip_max: Maximum value of an example.
        :param lbd: Weight for mask scoring function.
        :param pts: Optional. A set of points that will set the crop size in the perspective transform.
        :return: mask
        """
        # STAGE 3: Fine reduction
        theta = (x_tar_noise - x_noise) * mask

        # get initial transform_robustness
        xform_imgs = get_transformed_images(x, mask, xforms, 1.0, theta, self.net_size, clip_min, clip_max, pts)

        success_rate = run_predictions(self.estimator, xform_imgs, y, self.batch_size, False)

        init_tr_err = 1 - success_rate

        best_score = score_fn(mask, init_tr_err, object_size, threshold=(1 - self.tr_lo), lbd=lbd)
        best_mask = mask

        new_patches = []
        j = 0
        # iterate over patches, greedily remove if the score improves
        while patches:
            j = j + 1

            # highest transform_robustness is now always at end, for pop()
            next_patch = patches.pop()
            if np.max(next_patch * best_mask) == 0:
                continue
            next_mask = best_mask * (np.ones(best_mask.shape) - next_patch)
            theta = (x_tar_noise - x_noise) * next_mask

            xform_imgs = get_transformed_images(
                x, next_mask, xforms, 1.0, theta, self.net_size, clip_min, clip_max, pts
            )

            success_rate = run_predictions(self.estimator, xform_imgs, y, self.batch_size, False)

            score = score_fn(next_mask, 1 - success_rate, object_size, threshold=(1 - self.tr_lo), lbd=lbd)
            if score < best_score:
                best_score = score
                best_mask = next_mask
                nbits = best_mask.sum() / best_mask.shape[-1]
                if self.max_mask_size > 0 and nbits < self.max_mask_size:
                    break
            else:
                new_patches.append(next_patch)

        return best_mask

    def _boost(
        self,
        x: np.ndarray,
        x_noise: np.ndarray,
        x_tar_noise: np.ndarray,
        mask: np.ndarray,
        y: int,
        obj_width: float,
        focal: float,
        clip_min: float,
        clip_max: float,
        pts: Optional[np.ndarray] = None,
    ) -> np.ndarray:
        """
        Function to boost transform-robustness.

        :param x: An array with one original input to be attacked.
        :param x_noise: x in the resolution of the noise size.
        :param x_tar_noise: x_tar in the resolution of the noise size.
        :param mask: An array with a mask to be applied to the adversarial perturbations. Shape needs to be
                     broadcastable to the shape of x. Any features for which the mask is zero will not be adversarially
                     perturbed.
        :param y: The target label.
        :param obj_width: Estimated width of object in inches for perspective transform.
        :param focal: Estimated focal length in ft for perspective transform.
        :param clip_min: Minimum value of an example.
        :param clip_max: Maximum value of an example.
        :param pts: Optional. A set of points that will set the crop size in the perspective transform.
        :return: attacked image
        """

        xforms = get_transform_params(
            self.num_xforms_boost,
            self.rotation_range,
            self.dist_range,
            self.gamma_range,
            self.crop_percent_range,
            self.off_x_range,
            self.off_y_range,
            self.blur_kernels,
            obj_width,
            focal,
        )

        # Initialize
        theta = (x_tar_noise - x_noise) * mask
        query_count = 0

        xform_imgs = get_transformed_images(x, mask, xforms, 1.0, theta, self.net_size, clip_min, clip_max, pts)

        err_rate = run_predictions(self.estimator, xform_imgs, y, self.batch_size, True)
        query_count += self.num_xforms_boost

        best_theta, best_eps = theta, err_rate

        # End Initialize

        theta, eps = best_theta.copy(), best_eps

        opt_count = 0

        # gradient free optimization steps
        while True:
            gradient = np.zeros(theta.shape)
            num_q_samples = 10

            # Take q samples of random Gaussian noise to use as new directions, calculate transform_robustness
            # Used for gradient estimate
            for _ in range(num_q_samples):
                unit_dir = np.random.randn(*theta.shape).astype(np.float32) * mask
                unit_dir = unit_dir / np.linalg.norm(unit_dir)
                ttt = theta + self.beta * unit_dir

                xform_imgs = get_transformed_images(x, mask, xforms, 1.0, ttt, self.net_size, clip_min, clip_max, pts)
                eps_ttt = run_predictions(self.estimator, xform_imgs, y, self.batch_size, True)
                opt_count += self.num_xforms_boost

                gradient += (eps_ttt - eps) / self.beta * unit_dir

            gradient = 1.0 / num_q_samples * gradient

            # Take gradient step
            new_theta = theta - self.eta * gradient
            xform_imgs = get_transformed_images(x, mask, xforms, 1.0, new_theta, self.net_size, clip_min, clip_max, pts)
            new_eps = run_predictions(self.estimator, xform_imgs, y, self.batch_size, True)
            opt_count += self.num_xforms_boost

            if new_eps < best_eps:
                best_theta, best_eps = new_theta.copy(), new_eps

            theta, eps = new_theta.copy(), new_eps

            if (opt_count + query_count + self.num_xforms_boost * 11) > self.num_boost_queries:
                break

        adv_example, _, _ = add_noise(x, mask, 1.0, best_theta)

        return adv_example

    def _check_params(self) -> None:
        if self.noise_size < self.heat_patch_size:
            raise ValueError("Heatmap patch size must be smaller than the noise size.")

        if min(self.heat_patch_size) <= 0:
            raise ValueError("Heatmap patch size must be positive.")

        if not isinstance(self.heat_patch_size[0], int) or not isinstance(self.heat_patch_size[1], int):
            raise ValueError("Heatmap patch size must be a tuple of two integers.")

        if (
            min(self.heat_patch_stride) <= 0
            or not isinstance(self.heat_patch_stride[0], int)
            or not isinstance(self.heat_patch_stride[1], int)
        ):
            raise ValueError("Heatmap patch stride must be a tuple of two positive integers.")

        if self.heatmap_mode not in ["Target", "Random"]:
            raise ValueError("Heatmap mode must be 'Target' or 'Random'.")

        if self.tr_lo < 0 or self.tr_lo > 1:
            raise ValueError("tr_lo must be between 0 and 1.")

        if self.tr_hi < 0 or self.tr_hi > 1:
            raise ValueError("tr_hi must be between 0 and 1.")

        if self.tr_hi < self.tr_lo:
            raise ValueError("tr_hi must be at least as high as tr_lo.")

        if self.num_xforms_mask < 0 or not isinstance(self.num_xforms_mask, int):
            raise ValueError("num_xforms_mask must be non-negative integer.")

        if self.beta <= 0:
            raise ValueError("beta must be positive.")

        if self.eta <= 0:
            raise ValueError("eta must be positive.")

        if self.num_xforms_boost < 0 or not isinstance(self.num_xforms_boost, int):
            raise ValueError("num_xforms_boost must be positive integer.")

        if self.num_boost_queries <= 0 or not isinstance(self.num_boost_queries, int):
            raise ValueError("num_boost_queries must be positive.")

        if (
            self.rotation_range[0] <= -90
            or self.rotation_range[1] >= 90
            or self.rotation_range[1] < self.rotation_range[0]
        ):
            raise ValueError("rotation range must be within (-90, 90).")

        if self.dist_range[1] < self.dist_range[0] or self.dist_range[0] < 0:
            raise ValueError("distance range invalid. max must be greater than min, and must be nonnegative.")

        if self.gamma_range[1] < self.gamma_range[0] or self.gamma_range[0] < 1:
            raise ValueError("gamma range max must be greater than min and the range must be at 1.0 or greater.")

        if self.crop_percent_range[1] < self.crop_percent_range[0]:
            raise ValueError("max of crop percent range must be greater or equal to the min.")

        if self.off_x_range[1] < self.off_x_range[0]:
            raise ValueError("max of off x range must be greater or equal to the min.")

        if self.off_y_range[1] < self.off_y_range[0]:
            raise ValueError("max of off y range must be greater or equal to the min.")

        if min(self.blur_kernels) < 0:
            raise ValueError("blur kernels must be positive.")



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2022
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This module implements a metric for inference attack worst case accuracy measurement.
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import logging
from typing import Optional, List, Tuple, Union

import numpy as np
from sklearn.metrics import roc_curve


TPR = float  # True Positive Rate
FPR = float  # False Positive Rate
THR = float  # Threshold of the binary decision


def _calculate_roc_for_fpr(y_true: np.ndarray, y_proba: np.ndarray, targeted_fpr: float) -> Tuple[FPR, TPR, THR]:
    """
    Get FPR, TPR and, THRESHOLD based on the targeted_fpr (such that FPR <= targeted_fpr)
    :param y_true: True attack labels.
    :param y_proba: Predicted attack probabilities.
    :param targeted_fpr: the targeted False Positive Rate, ROC will be calculated based on this FPR.
    :return: tuple that contains (Achieved FPR, TPR, Threshold).
    """

    fpr, tpr, thr = roc_curve(y_true=y_true, y_score=y_proba)
    # take the highest fpr and an appropriated threshold that achieve at least FPR=fpr
    if np.isnan(fpr).all() or np.isnan(tpr).all():
        logging.error("TPR or FPR values are NaN")
        raise ValueError("The targeted FPR can't be achieved.")

    targeted_fpr_idx = np.where(fpr <= targeted_fpr)[0][-1]
    return fpr[targeted_fpr_idx], tpr[targeted_fpr_idx], thr[targeted_fpr_idx]


def get_roc_for_fpr(
    attack_proba: np.ndarray,
    attack_true: np.ndarray,
    target_model_labels: Optional[np.ndarray] = None,
    targeted_fpr: float = 0.001,
) -> Union[List[Tuple[FPR, TPR, THR]], List[Tuple[int, FPR, TPR, THR]]]:
    """
    Compute the attack TPR, THRESHOLD and achieved FPR based on the targeted FPR. This implementation supports only
    binary attack prediction labels {0,1}. The returned THRESHOLD defines the decision threshold on the attack
    probabilities (meaning if p < THRESHOLD predict 0, otherwise predict 1)
    | Related paper link: https://arxiv.org/abs/2112.03570

    :param attack_proba: Predicted attack probabilities.
    :param attack_true: True attack labels.
    :param targeted_fpr: the targeted False Positive Rate, attack accuracy will be calculated based on this FPRs.
     If not supplied, get_roc_for_fpr will be computed for `0.001` FPR.
    :param target_model_labels: Original labels, if provided the Accuracy and threshold will be calculated per each
     class separately.
    :return: list of tuples the contains (original label (if target_model_labels is provided),
    Achieved FPR, TPR, Threshold).
    """

    if attack_proba.shape[0] != attack_true.shape[0]:
        raise ValueError("Number of rows in attack_pred and attack_true do not match")
    if target_model_labels is not None and attack_proba.shape[0] != target_model_labels.shape[0]:
        raise ValueError("Number of rows in target_model_labels and attack_pred do not match")

    results = []

    if target_model_labels is not None:
        values, _ = np.unique(target_model_labels, return_counts=True)
        for value in values:
            idxs = np.where(target_model_labels == value)[0]
            fpr, tpr, thr = _calculate_roc_for_fpr(
                y_proba=attack_proba[idxs], y_true=attack_true[idxs], targeted_fpr=targeted_fpr
            )
            results.append((value, fpr, tpr, thr))
        return results

    fpr, tpr, thr = _calculate_roc_for_fpr(y_proba=attack_proba, y_true=attack_true, targeted_fpr=targeted_fpr)
    return [(fpr, tpr, thr)]


def get_roc_for_multi_fprs(
    attack_proba: np.ndarray,
    attack_true: np.ndarray,
    targeted_fprs: np.ndarray,
) -> Tuple[List[FPR], List[TPR], List[THR]]:
    """
    Compute the attack ROC based on the targeted FPRs. This implementation supports only binary
    attack prediction labels. The returned list of THRESHOLDs defines the decision threshold on the attack
    probabilities (meaning if p < THRESHOLD predict 0, otherwise predict 1) for each provided fpr

    | Related paper link: https://arxiv.org/abs/2112.03570

    :param attack_proba: Predicted attack probabilities.
    :param attack_true: True attack labels.
    :param targeted_fprs: the set of targeted FPR (False Positive Rates), attack accuracy will be calculated based on
    these FPRs.
    :return: list of tuples that  (TPR, Threshold, Achieved FPR).
    """

    if attack_proba.shape[0] != attack_true.shape[0]:
        raise ValueError("Number of rows in attack_pred and attack_true do not match")

    tpr = []
    thr = []
    fpr = []

    for t_fpr in targeted_fprs:
        res = _calculate_roc_for_fpr(y_proba=attack_proba, y_true=attack_true, targeted_fpr=t_fpr)

        fpr.append(res[0])
        tpr.append(res[1])
        thr.append(res[2])

    return fpr, tpr, thr



import logging

import numpy as np
import pytest
from jax import random
from jax import grad, jit, vmap
import jax.numpy as jnp
from jax.scipy.special import logsumexp

from art.experimental.estimators.classification.jax import JaxClassifier
from tests.utils import ARTTestException


logger = logging.getLogger(__name__)


# Define a jax model
def jax_classifier():
    # Define model sizes
    layer_sizes = [784, 256, 128, 10]

    # Initialize random weights
    def random_init(m, n, k):
        w, b = random.split(k)
        rw, rb = random.normal(w, (n, m)) * 0.01, random.normal(b, (n,)) * 0.01
        return rw, rb

    model = [
        random_init(m, n, k) for m, n, k in zip(layer_sizes[:-1], layer_sizes[1:], random.split(random.PRNGKey(0), 4))
    ]

    # Forward function
    def forward(model, x):
        activation = jnp.reshape(x, 784)
        for w, b in model[:-1]:
            output = jnp.dot(w, activation) + b
            activation = jnp.maximum(0, output)

        f_w, f_b = model[-1]
        logit = jnp.dot(f_w, activation) + f_b
        return logit - logsumexp(logit)

    # Prediction function
    predict_func = vmap(forward, in_axes=(None, 0))

    # Loss function
    def loss_func(model, x, y):
        preds = predict_func(model, x)
        return -jnp.mean(preds * y)

    # Update function
    @jit
    def update_func(model, x, y):
        grads = grad(loss_func)(model, x, y)
        return [(w - 0.01 * dw, b - 0.01 * db) for (w, b), (dw, db) in zip(model, grads)]

    classifier = JaxClassifier(
        model=model,
        predict_func=predict_func,
        loss_func=loss_func,
        update_func=update_func,
        input_shape=(28, 28, 1),
        nb_classes=10,
    )

    return classifier


classifier = jax_classifier()


@pytest.mark.skip_framework("pytorch", "tensorflow", "keras", "kerastf", "mxnet", "non_dl_frameworks")
def test_fit(art_warning, get_default_mnist_subset, default_batch_size):
    try:
        (x_train_mnist, y_train_mnist), (x_test_mnist, y_test_mnist) = get_default_mnist_subset

        labels = np.argmax(y_test_mnist, axis=1)

        accuracy = np.sum(np.argmax(classifier.predict(x_test_mnist), axis=1) == labels) / x_test_mnist.shape[0]
        np.testing.assert_array_almost_equal(accuracy, 0.32, decimal=0.06)

        classifier.fit(x_train_mnist, y_train_mnist, batch_size=default_batch_size, nb_epochs=2)
        accuracy_2 = np.sum(np.argmax(classifier.predict(x_test_mnist), axis=1) == labels) / x_test_mnist.shape[0]
        np.testing.assert_array_almost_equal(accuracy_2, 0.73, decimal=0.06)

    except ARTTestException as e:
        art_warning(e)


@pytest.mark.skip_framework("pytorch", "tensorflow", "keras", "kerastf", "mxnet", "non_dl_frameworks")
def test_predict(art_warning, get_default_mnist_subset, expected_values):
    try:
        (_, _), (x_test_mnist, y_test_mnist) = get_default_mnist_subset

        y_predicted = classifier.predict(x_test_mnist[0:1])

        np.testing.assert_array_almost_equal(y_predicted, expected_values(), decimal=4)

    except ARTTestException as e:
        art_warning(e)


@pytest.mark.skip_framework("pytorch", "tensorflow", "keras", "kerastf", "mxnet", "non_dl_frameworks")
def test_shapes(art_warning, get_default_mnist_subset):
    try:
        (_, _), (x_test_mnist, y_test_mnist) = get_default_mnist_subset

        predictions = classifier.predict(x_test_mnist)

        assert predictions.shape == y_test_mnist.shape

        assert classifier.nb_classes == 10

        loss_gradients = classifier.loss_gradient(x_test_mnist[:11], y_test_mnist[:11])

        assert loss_gradients.shape == x_test_mnist[:11].shape

    except ARTTestException as e:
        art_warning(e)


@pytest.mark.skip_framework("pytorch", "tensorflow", "keras", "kerastf", "mxnet", "non_dl_frameworks")
def test_loss_gradient(art_warning, get_default_mnist_subset, expected_values, mnist_shape):
    try:
        (expected_gradients_1, expected_gradients_2) = expected_values()

        (_, _), (x_test_mnist, y_test_mnist) = get_default_mnist_subset

        gradients = classifier.loss_gradient(x_test_mnist, y_test_mnist)

        assert gradients.shape == (x_test_mnist.shape[0],) + mnist_shape

        sub_gradients = gradients[0, 0, :, 0]
        np.testing.assert_array_almost_equal(
            sub_gradients,
            expected_gradients_1[0],
            decimal=expected_gradients_1[1],
        )

        sub_gradients = gradients[0, :, 14, 0]
        np.testing.assert_array_almost_equal(
            sub_gradients,
            expected_gradients_2[0],
            decimal=expected_gradients_2[1],
        )

    except ARTTestException as e:
        art_warning(e)


@pytest.mark.skip_framework("pytorch", "tensorflow", "keras", "kerastf", "mxnet", "non_dl_frameworks")
def test_nb_classes(art_warning):
    try:
        assert classifier.nb_classes == 10

    except ARTTestException as e:
        art_warning(e)


@pytest.mark.skip_framework("pytorch", "tensorflow", "keras", "kerastf", "mxnet", "non_dl_frameworks")
def test_input_shape(art_warning, mnist_shape):
    try:
        assert classifier.input_shape == mnist_shape

    except ARTTestException as e:
        art_warning(e)



"""
Module implementing train-based defences against adversarial attacks.
"""
from art.defences.trainer.trainer import Trainer
from art.defences.trainer.adversarial_trainer import AdversarialTrainer
from art.defences.trainer.certified_adversarial_trainer_pytorch import AdversarialTrainerCertifiedPytorch
from art.defences.trainer.ibp_certified_trainer_pytorch import AdversarialTrainerCertifiedIBPPyTorch
from art.defences.trainer.adversarial_trainer_madry_pgd import AdversarialTrainerMadryPGD
from art.defences.trainer.adversarial_trainer_fbf import AdversarialTrainerFBF
from art.defences.trainer.adversarial_trainer_fbf_pytorch import AdversarialTrainerFBFPyTorch
from art.defences.trainer.adversarial_trainer_trades import AdversarialTrainerTRADES
from art.defences.trainer.adversarial_trainer_trades_pytorch import AdversarialTrainerTRADESPyTorch
from art.defences.trainer.adversarial_trainer_awp import AdversarialTrainerAWP
from art.defences.trainer.adversarial_trainer_awp_pytorch import AdversarialTrainerAWPPyTorch
from art.defences.trainer.adversarial_trainer_oaat import AdversarialTrainerOAAT
from art.defences.trainer.adversarial_trainer_oaat_pytorch import AdversarialTrainerOAATPyTorch
from art.defences.trainer.dp_instahide_trainer import DPInstaHideTrainer



"""
This module implements all regressors in ART.
"""
from art.estimators.regression.regressor import RegressorMixin, Regressor

from art.estimators.regression.scikitlearn import ScikitlearnRegressor

from art.estimators.regression.keras import KerasRegressor

from art.estimators.regression.pytorch import PyTorchRegressor

from art.estimators.regression.blackbox import BlackBoxRegressor



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2023
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import absolute_import, division, print_function, unicode_literals

import logging

import numpy as np
import pytest

from tests.utils import ARTTestException

logger = logging.getLogger(__name__)


@pytest.mark.only_with_platform("pytorch")
def test_predict(art_warning, get_pytorch_detr):
    try:
        from art.utils import non_maximum_suppression

        object_detector, x_test, _ = get_pytorch_detr

        preds = object_detector.predict(x_test)
        result = non_maximum_suppression(preds[0], iou_threshold=0.4, confidence_threshold=0.3)
        assert list(result.keys()) == ["boxes", "labels", "scores"]

        assert result["boxes"].shape == (3, 4)
        expected_detection_boxes = np.asarray(
            [
                [1.0126123, 25.658852, 412.70746, 379.12537],
                [-0.089400, 272.08664, 415.90994, 416.25930],
                [0.1522941, 75.882440, 99.139565, 335.11273],
            ]
        )
        np.testing.assert_array_almost_equal(result["boxes"], expected_detection_boxes, decimal=3)

        assert result["scores"].shape == (3,)
        expected_detection_scores = np.asarray([0.8424455, 0.7796526, 0.35387915])
        np.testing.assert_array_almost_equal(result["scores"], expected_detection_scores, decimal=3)

        assert result["labels"].shape == (3,)
        expected_detection_classes = np.asarray([17, 65, 17])
        np.testing.assert_array_equal(result["labels"], expected_detection_classes)

    except ARTTestException as e:
        art_warning(e)


@pytest.mark.only_with_platform("pytorch")
def test_fit(art_warning, get_pytorch_detr):
    try:
        object_detector, x_test, y_test = get_pytorch_detr

        # Compute loss before training
        loss1 = object_detector.compute_loss(x=x_test, y=y_test)

        # Train for one epoch
        object_detector.fit(x_test, y_test, nb_epochs=1)

        # Compute loss after training
        loss2 = object_detector.compute_loss(x=x_test, y=y_test)

        assert loss1 != loss2

    except ARTTestException as e:
        art_warning(e)


@pytest.mark.only_with_platform("pytorch")
def test_loss_gradient(art_warning, get_pytorch_detr):
    try:
        object_detector, x_test, y_test = get_pytorch_detr

        grads = object_detector.loss_gradient(x=x_test, y=y_test)

        assert grads.shape == (1, 3, 416, 416)

        expected_gradients1 = np.asarray(
            [
                0.02891439,
                0.0055933,
                -0.00687808,
                0.0095074,
                0.00247894,
                0.00122704,
                -0.00482378,
                -0.00924361,
                -0.02870164,
                -0.00683936,
                0.00904205,
                -0.01315971,
                -0.0151937,
                -0.00156442,
                0.00775309,
                0.01946152,
                0.00523211,
                -0.01682214,
                0.00079588,
                0.01627164,
                -0.01347653,
                -0.00512358,
                0.00610363,
                0.02831643,
                0.00742467,
                0.00293561,
                0.01380033,
                0.02112359,
                0.01725711,
                -0.00431877,
                -0.01007722,
                -0.00526983,
            ]
        )
        np.testing.assert_array_almost_equal(grads[0, 0, 208, 192:224], expected_gradients1, decimal=2)

        expected_gradients2 = np.asarray(
            [
                -0.00549417,
                -0.01592844,
                -0.01073932,
                -0.00443333,
                -0.00780143,
                -0.02033146,
                -0.0191503,
                0.01227987,
                0.019971,
                0.01034214,
                -0.00918145,
                -0.02458049,
                -0.00708776,
                -0.00826812,
                -0.01284431,
                -0.00195021,
                0.00523211,
                0.00661678,
                0.00851441,
                0.01157211,
                -0.00324841,
                -0.00395823,
                0.00756641,
                0.00405913,
                -0.00055517,
                0.00221484,
                -0.02415526,
                -0.02096599,
                0.00980014,
                0.00174731,
                -0.01008899,
                0.00305779,
            ]
        )
        np.testing.assert_array_almost_equal(grads[0, 0, 192:224, 208], expected_gradients2, decimal=2)

    except ARTTestException as e:
        art_warning(e)


@pytest.mark.only_with_platform("pytorch")
def test_errors(art_warning):
    try:
        from torch import hub

        from art.estimators.object_detection.pytorch_detection_transformer import PyTorchDetectionTransformer

        model = hub.load("facebookresearch/detr", "detr_resnet50", pretrained=True)

        with pytest.raises(ValueError):
            PyTorchDetectionTransformer(
                model=model,
                clip_values=(1, 2),
                attack_losses=("loss_ce", "loss_bbox", "loss_giou"),
            )

        with pytest.raises(ValueError):
            PyTorchDetectionTransformer(
                model=model,
                clip_values=(-1, 1),
                attack_losses=("loss_ce", "loss_bbox", "loss_giou"),
            )

        from art.defences.postprocessor.rounded import Rounded

        post_def = Rounded()
        with pytest.raises(ValueError):
            PyTorchDetectionTransformer(
                model=model,
                clip_values=(0, 1),
                attack_losses=("loss_ce", "loss_bbox", "loss_giou"),
                postprocessing_defences=post_def,
            )

    except ARTTestException as e:
        art_warning(e)


@pytest.mark.only_with_platform("pytorch")
def test_preprocessing_defences(art_warning, get_pytorch_detr):
    try:
        object_detector, x_test, _ = get_pytorch_detr

        from art.defences.preprocessor.spatial_smoothing_pytorch import SpatialSmoothingPyTorch

        pre_def = SpatialSmoothingPyTorch()

        object_detector.set_params(preprocessing_defences=pre_def)

        # Create labels
        result = object_detector.predict(x=x_test)

        y = [
            {
                "boxes": result[0]["boxes"],
                "labels": result[0]["labels"],
                "scores": np.ones_like(result[0]["labels"]),
            }
        ]

        # Compute gradients
        grads = object_detector.loss_gradient(x=x_test, y=y)

        assert grads.shape == (1, 3, 416, 416)

    except ARTTestException as e:
        art_warning(e)


@pytest.mark.only_with_platform("pytorch")
def test_compute_losses(art_warning, get_pytorch_detr):
    try:
        object_detector, x_test, y_test = get_pytorch_detr
        losses = object_detector.compute_losses(x=x_test, y=y_test)
        assert len(losses) == 3

    except ARTTestException as e:
        art_warning(e)


@pytest.mark.only_with_platform("pytorch")
def test_compute_loss(art_warning, get_pytorch_detr):
    try:
        object_detector, x_test, y_test = get_pytorch_detr

        # Compute loss
        loss = object_detector.compute_loss(x=x_test, y=y_test)

        assert pytest.approx(2.172381, abs=0.1) == float(loss)

    except ARTTestException as e:
        art_warning(e)


@pytest.mark.only_with_platform("pytorch")
def test_pgd(art_warning, get_pytorch_detr):
    try:
        from art.attacks.evasion import ProjectedGradientDescent

        object_detector, x_test, y_test = get_pytorch_detr

        attack = ProjectedGradientDescent(estimator=object_detector, max_iter=2)
        x_test_adv = attack.generate(x=x_test, y=y_test)
        np.testing.assert_raises(AssertionError, np.testing.assert_array_equal, x_test_adv, x_test)

    except ARTTestException as e:
        art_warning(e)



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2022
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This module implements the task specific estimator for PyTorch YOLO v3 and v5 object detectors.

| Paper link: https://arxiv.org/abs/1804.02767
"""
import logging
from typing import List, Dict, Optional, Tuple, Union, TYPE_CHECKING

import numpy as np

from art.estimators.object_detection.pytorch_object_detector import PyTorchObjectDetector

if TYPE_CHECKING:
    # pylint: disable=C0412
    import torch

    from art.utils import CLIP_VALUES_TYPE, PREPROCESSING_TYPE
    from art.defences.preprocessor.preprocessor import Preprocessor
    from art.defences.postprocessor.postprocessor import Postprocessor

logger = logging.getLogger(__name__)


class PyTorchYolo(PyTorchObjectDetector):
    """
    This module implements the model- and task specific estimator for YOLO v3, v5 object detector models in PyTorch.

    | Paper link: https://arxiv.org/abs/1804.02767
    """

    def __init__(
        self,
        model: "torch.nn.Module",
        input_shape: Tuple[int, ...] = (3, 416, 416),
        optimizer: Optional["torch.optim.Optimizer"] = None,
        clip_values: Optional["CLIP_VALUES_TYPE"] = None,
        channels_first: bool = True,
        preprocessing_defences: Union["Preprocessor", List["Preprocessor"], None] = None,
        postprocessing_defences: Union["Postprocessor", List["Postprocessor"], None] = None,
        preprocessing: "PREPROCESSING_TYPE" = None,
        attack_losses: Tuple[str, ...] = (
            "loss_classifier",
            "loss_box_reg",
            "loss_objectness",
            "loss_rpn_box_reg",
        ),
        device_type: str = "gpu",
    ):
        """
        Initialization.

        :param model: YOLO v3 or v5 model wrapped as demonstrated in examples/get_started_yolo.py.
                      The output of the model is `List[Dict[str, torch.Tensor]]`, one for each input image.
                      The fields of the Dict are as follows:

                      - boxes [N, 4]: the boxes in [x1, y1, x2, y2] format, with 0 <= x1 < x2 <= W and
                        0 <= y1 < y2 <= H.
                      - labels [N]: the labels for each image.
                      - scores [N]: the scores of each prediction.
        :param input_shape: The shape of one input sample.
        :param optimizer: The optimizer for training the classifier.
        :param clip_values: Tuple of the form `(min, max)` of floats or `np.ndarray` representing the minimum and
               maximum values allowed for features. If floats are provided, these will be used as the range of all
               features. If arrays are provided, each value will be considered the bound for a feature, thus
               the shape of clip values needs to match the total number of features.
        :param channels_first: Set channels first or last.
        :param preprocessing_defences: Preprocessing defence(s) to be applied by the classifier.
        :param postprocessing_defences: Postprocessing defence(s) to be applied by the classifier.
        :param preprocessing: Tuple of the form `(subtrahend, divisor)` of floats or `np.ndarray` of values to be
               used for data preprocessing. The first value will be subtracted from the input. The input will then
               be divided by the second one.
        :param attack_losses: Tuple of any combination of strings of loss components: 'loss_classifier', 'loss_box_reg',
                              'loss_objectness', and 'loss_rpn_box_reg'.
        :param device_type: Type of device to be used for model and tensors, if `cpu` run on CPU, if `gpu` run on GPU
                            if available otherwise run on CPU.
        """
        super().__init__(
            model=model,
            input_shape=input_shape,
            optimizer=optimizer,
            clip_values=clip_values,
            channels_first=channels_first,
            preprocessing_defences=preprocessing_defences,
            postprocessing_defences=postprocessing_defences,
            preprocessing=preprocessing,
            attack_losses=attack_losses,
            device_type=device_type,
        )

    def _translate_labels(self, labels: List[Dict[str, "torch.Tensor"]]) -> "torch.Tensor":
        """
        Translate object detection labels from ART format (torchvision) to the model format (YOLO) and
        move tensors to GPU, if applicable.

        :param labels: Object detection labels in format x1y1x2y2 (torchvision).
        :return: Object detection labels in format xcycwh (YOLO).
        """
        import torch

        if self.channels_first:
            height = self.input_shape[1]
            width = self.input_shape[2]
        else:
            height = self.input_shape[0]
            width = self.input_shape[1]

        labels_xcycwh_list = []

        for i, label_dict in enumerate(labels):
            # create 2D tensor to encode labels and bounding boxes
            label_xcycwh = torch.zeros(len(label_dict["boxes"]), 6, device=self.device)
            label_xcycwh[:, 0] = i
            label_xcycwh[:, 1] = label_dict["labels"]
            label_xcycwh[:, 2:6] = label_dict["boxes"]

            # normalize bounding boxes to [0, 1]
            label_xcycwh[:, 2:6:2] /= width
            label_xcycwh[:, 3:6:2] /= height

            # convert from x1y1x2y2 to xcycwh
            label_xcycwh[:, 4] -= label_xcycwh[:, 2]
            label_xcycwh[:, 5] -= label_xcycwh[:, 3]
            label_xcycwh[:, 2] += label_xcycwh[:, 4] / 2
            label_xcycwh[:, 3] += label_xcycwh[:, 5] / 2
            labels_xcycwh_list.append(label_xcycwh)

        labels_xcycwh = torch.vstack(labels_xcycwh_list)
        return labels_xcycwh

    def _translate_predictions(self, predictions: "torch.Tensor") -> List[Dict[str, np.ndarray]]:
        """
        Translate object detection predictions from the model format (YOLO) to ART format (torchvision) and
        convert tensors to numpy arrays.

        :param predictions: Object detection labels in format xcycwh (YOLO).
        :return: Object detection labels in format x1y1x2y2 (torchvision).
        """
        import torch

        if self.channels_first:
            height = self.input_shape[1]
            width = self.input_shape[2]
        else:
            height = self.input_shape[0]
            width = self.input_shape[1]

        predictions_x1y1x2y2: List[Dict[str, np.ndarray]] = []

        for pred in predictions:
            boxes = torch.vstack(
                [
                    torch.maximum((pred[:, 0] - pred[:, 2] / 2), torch.tensor(0, device=self.device)),
                    torch.maximum((pred[:, 1] - pred[:, 3] / 2), torch.tensor(0, device=self.device)),
                    torch.minimum((pred[:, 0] + pred[:, 2] / 2), torch.tensor(height, device=self.device)),
                    torch.minimum((pred[:, 1] + pred[:, 3] / 2), torch.tensor(width, device=self.device)),
                ]
            ).permute((1, 0))
            labels = torch.argmax(pred[:, 5:], dim=1)
            scores = pred[:, 4]

            pred_dict = {
                "boxes": boxes.detach().cpu().numpy(),
                "labels": labels.detach().cpu().numpy(),
                "scores": scores.detach().cpu().numpy(),
            }

            predictions_x1y1x2y2.append(pred_dict)

        return predictions_x1y1x2y2



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2020
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This module implements membership inference attacks.
"""

from __future__ import absolute_import, division, print_function, unicode_literals

import logging
from typing import Optional, TYPE_CHECKING

import numpy as np

from art.attacks.attack import MembershipInferenceAttack
from art.estimators.estimator import BaseEstimator
from art.estimators.classification.classifier import ClassifierMixin
from art.utils import check_and_transform_label_format

if TYPE_CHECKING:
    from art.utils import CLASSIFIER_TYPE

logger = logging.getLogger(__name__)


class MembershipInferenceBlackBoxRuleBased(MembershipInferenceAttack):
    """
    Implementation of a simple, rule-based black-box membership inference attack.

    This implementation uses the simple rule: if the model's prediction for a sample is correct, then it is a
    member. Otherwise, it is not a member.
    """

    attack_params = MembershipInferenceAttack.attack_params
    _estimator_requirements = (BaseEstimator, ClassifierMixin)

    def __init__(self, classifier: "CLASSIFIER_TYPE"):
        """
        Create a MembershipInferenceBlackBoxRuleBased attack instance.

        :param classifier: Target classifier.
        """
        super().__init__(estimator=classifier)

    def infer(self, x: np.ndarray, y: Optional[np.ndarray] = None, **kwargs) -> np.ndarray:
        """
        Infer membership in the training set of the target estimator.

        :param x: Input records to attack.
        :param y: True labels for `x`.
        :param probabilities: a boolean indicating whether to return the predicted probabilities per class, or just
                              the predicted class.
        :return: An array holding the inferred membership status, 1 indicates a member and 0 indicates non-member,
                 or class probabilities.
        """
        if y is None:  # pragma: no cover
            raise ValueError("MembershipInferenceBlackBoxRuleBased requires true labels `y`.")

        if self.estimator.input_shape is not None:  # pragma: no cover
            if self.estimator.input_shape[0] != x.shape[1]:
                raise ValueError("Shape of x does not match input_shape of classifier")

        if "probabilities" in kwargs:
            probabilities = kwargs.get("probabilities")
        else:
            probabilities = False

        y = check_and_transform_label_format(y, nb_classes=len(np.unique(y)), return_one_hot=True)
        if y is None:
            raise ValueError("None value detected.")
        if y.shape[0] != x.shape[0]:  # pragma: no cover
            raise ValueError("Number of rows in x and y do not match")

        # get model's predictions for x
        y_pred = self.estimator.predict(x=x)
        predicted_class = (np.argmax(y, axis=1) == np.argmax(y_pred, axis=1)).astype(int)
        if probabilities:
            # use y_pred as the probability if binary classification, otherwise just use 1
            if y_pred.shape[1] == 2:
                pred_prob = np.max(y_pred, axis=1)
                prob = np.zeros((predicted_class.shape[0], 2))
                prob[:, predicted_class] = pred_prob
                prob[:, np.ones_like(predicted_class) - predicted_class] = np.ones_like(pred_prob) - pred_prob
            else:
                # simply returns probability 1 for the predicted class and 0 for the other class
                prob_none = check_and_transform_label_format(predicted_class, nb_classes=2, return_one_hot=True)
                if prob_none is not None:
                    prob = prob_none
            return prob
        return predicted_class
